###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# toolbox/                                                                    #
#         cuckoo.py                                                           #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2022 science + computing ag                              #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################

""" Interface to Cuckoo. """

import datetime
import os
import logging
import threading

import requests
import schema
import urllib3.util.retry

from peekaboo.exceptions import PeekabooException


logger = logging.getLogger(__name__)


class CuckooSubmitFailedException(PeekabooException):
    """ An exception raised if submitting a job to Cuckoo fails. """
    pass


class CuckooJob:
    """ Remember sample and submission time of a Cuckoo job. """
    def __init__(self, sample):
        self.__sample = sample
        self.__submission_time = datetime.datetime.utcnow()

    def is_older_than(self, seconds):
        """ Returns True if the difference between submission time and now,
        i.e. the age of the job, is larger than given number of seconds. """
        max_age = datetime.timedelta(seconds=seconds)
        return datetime.datetime.utcnow() - self.__submission_time > max_age

    @property
    def sample(self):
        """ Returns the sample the job is analyzing. """
        return self.__sample


class AllowedStatusRetry(urllib3.util.retry.Retry):
    """ A Retry class which has a list of allowed status codes, allowing to
    retry all status codes not explicitly allowed in a hard-core, catch-all
    manner. """
    def __init__(self, allowed_statuses=None, abort=None, **kwargs):
        super().__init__(**kwargs)
        self.allowed_statuses = allowed_statuses or set()
        # Event that is set if we're not to retry
        self.abort = abort

    def new(self, **kwargs):
        """ Adjusted shallow copy method to carry our parameters over into our
        copy. """
        if 'allowed_statuses' not in kwargs:
            kwargs['allowed_statuses'] = self.allowed_statuses
        if 'abort' not in kwargs:
            kwargs['abort'] = self.abort
        return super().new(**kwargs)

    def is_exhausted(self):
        """ Allow to abort a retry chain through an external signal. """
        if self.abort and self.abort.is_set():
            return True

        return super().is_exhausted()

    def is_retry(self, method, status_code, has_retry_after=False):
        """ Override Retry's is_retry to introduce our allowed statuses logic.
        """
        # we retry all methods so no check if method is retryable here

        if self.allowed_statuses and status_code not in self.allowed_statuses:
            return True

        return super().is_retry(method, status_code, has_retry_after)


class Cuckoo:
    """ Interfaces with a Cuckoo installation via its REST API. """
    def __init__(self, job_queue, url="http://localhost:8090", api_token="",
                 poll_interval=5, submit_original_filename=True,
                 max_job_age=900, retries=5, backoff=0.5):
        """ Initialize the object.

        @param job_queue: The job queue to use from now on
        @type job_queue: JobQueue object
        @param url: Where to reach the Cuckoo REST API
        @type url: string
        @param api_token: API token to use for authentication
        @type api_token: string
        @param poll_interval: How long to wait inbetween job status checks
        @type poll_interval: int
        @param submit_original_filename: Whether to provide the original
                                         filename to Cuckoo to enhance analysis.
        @type submit_original_filename: bool
        @param max_job_age: How long to track jobs before declaring them failed.
        @type max_job_age: int (seconds)
        @param retries: Number of retries on API requests
        @type retries: int
        @param backoff: Backoff factor for urllib3
        @type backoff: float
        """
        self.job_queue = job_queue
        self.shutdown_requested = threading.Event()
        self.shutdown_requested.clear()
        self.running_jobs = {}
        # reentrant because we're doing nested calls within critical sections
        self.running_jobs_lock = threading.RLock()
        self.url = url
        self.api_token = api_token
        self.poll_interval = poll_interval
        self.submit_original_filename = submit_original_filename
        self.max_job_age = max_job_age

        # urrlib3 backoff formula:
        # <backoff factor> * (2 ^ (<retry count so far> - 1))
        # with second try intentionally having no sleep,
        # e.g. with retry count==5 and backoff factor==0.5:
        # try 1: fail, sleep(0.5*2^(1-1)==0.5*2^0==0.5*1==0.5->intentionally
        #   overridden to 0)
        # try 2: fail, sleep(0.5*2^(2-1)==0.5*2^1==1)
        # try 3: fail, sleep(0.5*2^(3-1)==0.5*2^2==2)
        # try 4: fail, sleep(0.5*2^(4-1)==0.5*2^3==4)
        # try 5: fail, abort, sleep would've been 8 before try 6
        #
        # Also, use allowed_methods=None to enable POST and other methods for
        # retry which aren't by default because they're not considered
        # idempotent. We assume that with the REST API a request either
        # succeeds or fails without residual effects, making them atomic and
        # idempotent.
        #
        # And finally we retry everything but a 200 response, which admittedly
        # is a bit hard-core but serves our purposes for now.
        retry_config = AllowedStatusRetry(
            total=retries, backoff_factor=backoff, allowed_methods=None,
            allowed_statuses=set([200]), abort=self.shutdown_requested)
        retry_adapter = requests.adapters.HTTPAdapter(max_retries=retry_config)
        self.session = requests.session()
        self.session.mount('http://', retry_adapter)
        self.session.mount('https://', retry_adapter)

        self.tracker = None

    def request_url(self, path):
        """ Return the full request URL for a given path based on the
        configured base URL of the Cuckoo API.

        @param path: resource path relative to the API base URL
        @type path: string

        @returns: string representing URL
        """
        return "%s/%s" % (self.url, path)

    def __get(self, path):
        request_url = self.request_url(path)
        logger.debug("Getting %s", request_url)
        headers = {"Authorization": "Bearer %s" % self.api_token}

        try:
            response = self.session.get(request_url, headers=headers)
        # all requests exceptions are derived from RequestsException, including
        # RetryError, TooManyRedirects and Timeout
        except requests.exceptions.RequestException as error:
            logger.error('Request to REST API failed: %s', error)
            return None

        # no check for status code here since we retry all but 200
        # responses and raise an exception if retries fail
        try:
            json_resp = response.json()
        except ValueError as error:
            logger.error(
                'Invalid JSON in response when getting %s: %s',
                request_url, error)
            return None

        return json_resp

    def register_running_job(self, job_id, sample):
        """ Register a job as running. Detect if another sample has already
        been registered with the same job ID which obviously must never happen
        because it corrupts our internal housekeeping. Guarded by a lock
        because multiple worker threads will call this routine and check for
        collision and update of job log might otherwise race each other.

        @param job_id: ID of the job to register as running.
        @type job_id: int
        @param sample: Sample object to associate with this job ID
        @type sample: Sample

        @returns: None
        @raises: CuckooSubmitFailedException on job id collision """
        with self.running_jobs_lock:
            if (job_id in self.running_jobs and
                    self.running_jobs[job_id] is not sample):
                logger.warning(
                    '%d: A job with ID %d already registered as running '
                    'for different sample %d will be marked failed',
                    sample.id, job_id,
                    self.running_jobs[job_id].sample.id)
                self.resubmit_as_failed(job_id)

            self.running_jobs[job_id] = CuckooJob(sample)

    def deregister_running_job(self, job_id):
        """ Deregister a running job by job id.

        @returns: Sample object of the job or None if job not found. """
        with self.running_jobs_lock:
            job = self.running_jobs.pop(job_id, None)
            if job is not None:
                return job.sample

        return None

    def deregister_running_job_if_too_old(self, job_id, max_age):
        """ Check if a job has gotten too old and remove it from the list of
        running jobs if so.

        @returns: Sample object of the job or None if job not found. """
        with self.running_jobs_lock:
            if self.running_jobs[job_id].is_older_than(max_age):
                return self.deregister_running_job(job_id)

        return None

    def resubmit_with_report(self, job_id):
        """ Resubmit a sample to the job queue after the report became
        available. Retrieves the report from Cuckoo.

        @param job_id: ID of job which has finished.
        @type job_id: int

        @returns: None """
        logger.debug("Analysis done for task #%d", job_id)

        sample = self.deregister_running_job(job_id)
        if sample is None:
            logger.debug('No sample found for job ID %d', job_id)
            return None

        logger.debug('%d: Requesting Cuckoo report', sample.id)
        report_path = "tasks/report/%d" % job_id
        report = self.__get(report_path)
        if report is None:
            # mark analysis as failed if we could not get the report e.g.
            # because it was corrupted or the API connection failed.
            sample.mark_cuckoo_failure()
        else:
            try:
                reportobj = CuckooReport(report, self.request_url(report_path))
                sample.register_cuckoo_report(reportobj)
            except schema.SchemaError as err:
                logger.warning('Report returned from Cuckoo contained '
                               'invalid data: %s', err)
                sample.mark_cuckoo_failure()

        self.job_queue.submit(sample)
        return None

    def resubmit_as_failed(self, job_id):
        """ Resubmit a sample to the job queue with a failure report if the
        Cuckoo job has failed for some reason.

        @param job_id: ID of job that failed.
        @type job_id: int
        """
        sample = self.deregister_running_job(job_id)
        if sample is not None:
            sample.mark_cuckoo_failure()
            self.job_queue.submit(sample)

    def resubmit_as_failed_if_too_old(self, job_id, max_age):
        """ Resubmit a sample to the job queue with a failure report if the
        Cuckoo job has been running for too long.

        @param job_id: ID of job to check.
        @type job_id: int
        @param max_age: maximum job age in seconds
        @type max_age: int
        """
        sample = self.deregister_running_job_if_too_old(job_id, max_age)
        if sample is not None:
            logger.warning("Dropped job %d because it has been running for "
                           "too long", job_id)
            sample.mark_cuckoo_failure()
            self.job_queue.submit(sample)

    def submit(self, sample):
        """ Submit a sample to Cuckoo for analysis.
        @param sample: Sample to submit.
        @type sample: Sample

        @raises: CuckooSubmitFailedException if submission failed
        @returns: ID of the submitted Cuckoo job.
        """
        filename = sample.sha256sum

        # append file extension to aid cuckoo in file type detection
        if sample.file_extension:
            filename = '%s.%s' % (filename, sample.file_extension)

        # override with the original file name if available
        if self.submit_original_filename and sample.filename:
            filename = sample.filename

        files = {"file": (filename, sample.content)}
        logger.debug("Creating Cuckoo task with filename %s", filename)
        headers = {"Authorization": "Bearer %s" % self.api_token}

        try:
            response = self.session.post(
                "%s/tasks/create/file" % self.url,
                headers=headers, files=files)
        except requests.exceptions.RequestException as error:
            raise CuckooSubmitFailedException(
                'Error creating Cuckoo task: %s' % error)

        try:
            json_resp = response.json()
        except ValueError as error:
            raise CuckooSubmitFailedException(
                'Invalid JSON in response when creating Cuckoo task: %s'
                % error)

        if "task_id" not in json_resp:
            raise CuckooSubmitFailedException(
                'No job ID present in API response')

        task_id = json_resp["task_id"]
        if not isinstance(task_id, int):
            raise CuckooSubmitFailedException(
                'Invalid data type for job ID in API response')

        if task_id is None or task_id <= 0:
            raise CuckooSubmitFailedException(
                'Invalid value for job ID in API response')

        self.register_running_job(task_id, sample)
        return task_id

    def start_tracker(self):
        """ Start tracking running jobs in a separate thread. """
        # do a simple initial test of connectivity. With this we require the
        # API to be reachable at startup (with the usual retries to account for
        # a bit of a race condition in parallel startup) but later on hope
        # that all errors are transient and retry endlessly
        status = self.__get("cuckoo/status")
        if status is None:
            logger.critical("Connection to Cuckoo REST API failed")
            return False
        if "tasks" not in status or "reported" not in status["tasks"]:
            logger.critical("Invalid status JSON structure from Cuckoo REST "
                            "API: %s", status)
            return False

        reported = status["tasks"]["reported"]
        logger.info("Connection to Cuckoo seems to work, "
                    "%i reported tasks seen", reported)

        self.tracker = threading.Thread(target=self.track,
                                        name="CuckooJobTracker")
        self.tracker.start()
        return True

    def track(self):
        """ Do the polling for finished jobs. """
        while not self.shutdown_requested.wait(self.poll_interval):
            # no lock, atomic, copy() because keys() returns an iterable view
            # instead of a fresh new list in python3
            running_jobs = self.running_jobs.copy().keys()

            # somewhat inefficient for the potential number of requests per
            # poll round but more efficient in that it does not download an
            # ever growing list of jobs using tasks/list. Might also take
            # quite a while to get to all jobs if retries happen on each
            # request.
            # A call to get data about multiple tasks in one go would be nice
            # here. tasks/list could be used with the minimum job number as
            # offset and spread between highest and lowest job id as limit *if*
            # its output was sorted by job ID. Apparently limit and offset are
            # only meant to iterate over the job list in blocks but not to
            # return data about a specific range of job IDs from that list.
            for job_id in running_jobs:
                job = self.__get("tasks/view/%i" % job_id)
                if job is None:
                    # ignore and retry on next polling run
                    continue

                if "task" not in job or "status" not in job["task"]:
                    logger.error("Invalid JSON structure from Cuckoo REST "
                                 "API: %s", job)
                    continue

                if job["task"]["status"] == "reported":
                    self.resubmit_with_report(job_id)
                    continue

                # drop jobs which have been running for too long. This is
                # mainly to prevent accumulation of jobs in our job list which
                # will never finish. We still want to wait for jobs to finish
                # even though our client might not be interested any more so
                # that we have the result cached for the next time we get the
                # same sample.
                self.resubmit_as_failed_if_too_old(job_id, self.max_job_age)

        logger.debug("Cuckoo job tracker shut down.")

    def shut_down(self):
        """ Request the module to shut down, used by the signal handler. """
        logger.debug("Cuckoo job tracker shutdown requested.")
        self.shutdown_requested.set()

    def close_down(self):
        """ Close down tracker resources, particularly, wait for the thread to
        terminate. """
        if self.tracker:
            self.tracker.join()
            self.tracker = None


class CuckooReport:
    """ Represents a Cuckoo analysis JSON report. """
    def __init__(self, report=None, url=None):
        """
        @param report: hash with report data from Cuckoo
        @type report: dict
        @param url: URL where the report was retrieved from
        @type url: string
        """
        self._url = url

        if report is None:
            report = {}

        # some common building blocks for reuse
        dns_element_schema = {'request': str}
        description_element_schema = {'description': str}

        # defaults of optional keys are not validated. Therefore their
        # validators can't set more default values. So we can only rely on the
        # validation result to contain the top-level key defaults. To avoid
        # confusion make no assumptions about optional key existance at all and
        # only schema compliance. We still use the result though because
        # ignore_extra_keys has stripped it of extraneous data which protects
        # us somewhat from accidentally processing it.
        report = schema.Schema({
            schema.Optional('network', default={}, ignore_extra_keys=True): {
                schema.Optional('dns', default=[]): schema.Or(
                    list([dns_element_schema]),
                    tuple([dns_element_schema]),
                    ignore_extra_keys=True),
                },
            schema.Optional('signatures', default=[]): schema.Or(
                list([description_element_schema]),
                tuple([description_element_schema]),
                ignore_extra_keys=True),
            schema.Optional('info', default={}): {
                schema.Optional('score', default=0.0): schema.Or(int, float),
                },
            schema.Optional('debug', default={}): {
                schema.Optional('errors', default=[]): schema.Or(
                    list([str]),
                    tuple([str])),
                schema.Optional('cuckoo', default=[]): schema.Or(
                    list([str]),
                    tuple([str])),
                },
            }, ignore_extra_keys=True).validate(report)

        self._requested_domains = [
            domain['request'] for domain in report.get(
                'network', {}).get('dns', [])]

        self._signature_descriptions = [
            sig['description'] for sig in report.get('signatures', [])]

        # explicitly convert to the types of our external API here if we accept
        # multiple types as input (schema.Use could convert as well but does it
        # before validation in duck-typing fashion which could make us accept
        # unintended types, e.g. a string because it can be converted to a list
        # because it's iterable).
        self._score = float(report.get('info', {}).get('score', 0.0))

        debug = report.get('debug', {})
        self._errors = list(debug.get('errors', []))
        self._server_messages = list(debug.get('cuckoo', []))

    @property
    def dump(self):
        """ Return the a dump of the report in a defined structure similar to
        the original Cuckoo report dict.

        @returns: dict containiing all the information we have.
        """
        return {
            "x-peekaboo": {
                "origin-url": self.url,
            },
            "info": {
                "score": self.score,
            },
            "network": {
                "dns": [
                    {"request": domain} for domain in self.requested_domains],
            },
            "signatures": [
                {"description": desc} for desc in self.signature_descriptions],
            "debug": {
                "errors": self.errors,
                "cuckoo": self.server_messages,
            },
        }

    @property
    def requested_domains(self):
        """
        Gets the requested domains from the Cuckoo report.

        @returns: The requested domains from the Cuckoo report.
        """
        return self._requested_domains

    @property
    def signature_descriptions(self):
        """
        Gets the description of triggered Cuckoo signatures from report.

        @returns: The description of triggered signatures from the Cuckoo
                  report.
        """
        return self._signature_descriptions

    @property
    def score(self):
        """
        Gets the score from the Cuckoo report.

        @returns: The score from the Cuckoo report.
        """
        return self._score

    @property
    def errors(self):
        """
        Errors occurred during Cuckoo analysis.

        @returns: The errors occurred during Cuckoo analysis.
        """
        return self._errors

    @property
    def server_messages(self):
        """
        Messages logged by the Cuckoo server (as opposed to those logged by the
        agent inside the analysis VM).

        @returns: List of messages.
        """
        return self._server_messages

    @property
    def url(self):
        """ Return a URL where the full Cuckoo report this object is based on
        can be retrieved (again).

        @returns: string representing the URL.
        """
        return self._url



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# queuing.py                                                                  #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2022 science + computing ag                              #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################

""" The main job queue and worker threads. """


import logging
import queue
import threading
import time

from peekaboo.ruleset import Result, RuleResult
from peekaboo.ruleset.engine import RulesetEngine
from peekaboo.exceptions import (
    PeekabooAnalysisDeferred, PeekabooDatabaseError, PeekabooConfigException,
    PeekabooRulesetConfigError)

logger = logging.getLogger(__name__)


class JobQueue:
    """ Peekaboo's queuing system. """
    def __init__(self, ruleset_config, db_con, analyzer_config, worker_count=4,
                 queue_timeout=300, shutdown_timeout=60,
                 cluster_duplicate_check_interval=5):
        """ Initialise job queue by creating n Peekaboo worker threads to
        process samples.

        @param ruleset_config: the ruleset configuration
        @type ruleset_config: PeekabooConfigParser
        @param db_con: Database connection object for cluster instance
                       coordination, i.e. saving sample info.
        @type db_con: PeekabooDatabase
        @param worker_count: The amount of worker threads to create. Defaults
                             to 4.
        @type worker_count: int
        @param queue_timeout: How long to block before considering queueing
                              failed.
        @type queue_timeout: int
        @param shutdown_timeout: How long to block before considering shutdown
                                 failed.
        @type shutdown_timeout: int
        @param cluster_duplicate_check_interval: How long to wait inbetween
                                                 checks for stale cluster
                                                 duplicate locks.
        @type cluster_duplicate_check_interval: int
        @raises PeekabooConfigException: if an error occured in configuration.
        """
        self.db_con = db_con
        self.jobs = queue.Queue()
        self.workers = []
        self.worker_count = worker_count
        self.queue_timeout = queue_timeout
        self.shutdown_timeout = shutdown_timeout

        # keep a backlog of samples with hashes identical to samples currently
        # in analysis to avoid analysing multiple identical samples
        # simultaneously. Once one analysis has finished, we can submit the
        # others and the ruleset will notice that we already know the result.
        self.duplicates = {}
        self.duplock = threading.Lock()

        # keep a similar backlog of samples currently being processed by
        # other instances so we can regularly try to resubmit them and re-use
        # the other instances' cached results from the database
        self.cluster_duplicates = {}

        self.ruleset_engine = RulesetEngine(
            ruleset_config, self, db_con, analyzer_config)

        # we start these here because they do no lengthy init and starting can
        # not fail. We need this here to avoid races in startup vs. shutdown by
        # signal to avoid continuing running in a half-inited/half-shutdown
        # state.
        for wno in range(0, self.worker_count):
            logger.debug("Create Worker %d", wno)
            worker = Worker(wno, self, self.ruleset_engine, db_con)
            self.workers.append(worker)

        logger.info('Created %d Workers.', self.worker_count)

        self.cluster_duplicate_handler = None
        if cluster_duplicate_check_interval:
            logger.debug(
                "Starting cluster duplicate handler thread with check "
                "interval %d.", cluster_duplicate_check_interval)
            self.cluster_duplicate_handler = ClusterDuplicateHandler(
                self, cluster_duplicate_check_interval)
        else:
            logger.debug("Disabling cluster duplicate handler thread.")

    def start(self):
        """ Start up the job queue including resource initialisation. """
        for worker in self.workers:
            worker.start()

        if self.cluster_duplicate_handler:
            self.cluster_duplicate_handler.start()

        # create a single ruleset engine for all workers, instantiates all the
        # rules based on the ruleset configuration, may start up long-lived
        # analyzer instances which are shared as well, is otherwise stateless
        # to allow concurrent use by multiple worker threads
        try:
            self.ruleset_engine.start()
        except (KeyError, ValueError, PeekabooConfigException) as error:
            self.shut_down()
            self.close_down()
            raise PeekabooConfigException(
                'Ruleset configuration error: %s' % error)
        except PeekabooRulesetConfigError as error:
            self.shut_down()
            self.close_down()
            raise PeekabooConfigException(error)

    def submit(self, sample):
        """
        Adds a Sample object to the job queue.
        If the queue is full, we block for 300 seconds and then throw an
        exception.

        @param sample: The Sample object to add to the queue.
        @raises Full: if the queue is full.
        """
        sample_hash = sample.sha256sum
        duplicate = None
        cluster_duplicate = None
        resubmit = None
        # we have to lock this down because apart from callbacks from our
        # Workers we're also called from the ThreadingUnixStreamServer
        with self.duplock:
            # check if a sample with same hash is currently in flight
            duplicates = self.duplicates.get(sample_hash)
            if duplicates is not None:
                # we are regularly resubmitting samples, e.g. after we've
                # noticed that cuckoo is finished analysing them. This
                # obviously isn't a duplicate but continued processing of the
                # same sample.
                if duplicates['master'] == sample:
                    resubmit = sample.id
                    self.jobs.put(sample, True, self.queue_timeout)
                else:
                    # record the to-be-submitted sample as duplicate and do
                    # nothing
                    duplicate = sample.id
                    duplicates['duplicates'].append(sample)
            else:
                # are we the first of potentially multiple instances working on
                # this sample?
                try:
                    locked = self.db_con.mark_sample_in_flight(sample)
                except PeekabooDatabaseError as dberr:
                    logger.error(dberr)
                    return False

                if locked:
                    # initialise a per-duplicate backlog for this sample which
                    # also serves as in-flight marker and submit to queue
                    self.duplicates[sample_hash] = {
                        'master': sample,
                        'duplicates': [],
                    }
                    self.jobs.put(sample, True, self.queue_timeout)
                else:
                    # another instance is working on this
                    if self.cluster_duplicates.get(sample_hash) is None:
                        self.cluster_duplicates[sample_hash] = []

                    cluster_duplicate = sample.id
                    self.cluster_duplicates[sample_hash].append(sample)

        if duplicate is not None:
            logger.debug(
                "%d: Sample is duplicate and waiting for running analysis "
                "to finish", duplicate)
        elif cluster_duplicate is not None:
            logger.debug(
                "%d: Sample is concurrently processed by another instance "
                "and held", cluster_duplicate)
        elif resubmit is not None:
            logger.debug("%d: Resubmitted sample to job queue", resubmit)
        else:
            logger.debug("%d: New sample submitted to job queue", sample.id)

        return True

    def submit_cluster_duplicates(self):
        """ Submit samples held while being processed by another cluster
        instance back into the job queue if they have finished processing. """
        if not self.cluster_duplicates.keys():
            return True

        submitted_cluster_duplicates = []

        with self.duplock:
            # try to submit *all* samples which have been marked as being
            # processed by another instance concurrently
            # get the items view on a copy of the cluster duplicate backlog
            # because we will change it by removing entries which would raise a
            # RuntimeException
            cluster_duplicates = self.cluster_duplicates.copy().items()
            for sample_hash, sample_duplicates in cluster_duplicates:
                # try to mark as in-flight
                try:
                    locked = self.db_con.mark_sample_in_flight(
                        sample_duplicates[0])
                except PeekabooDatabaseError as dberr:
                    logger.error(dberr)
                    return False

                if locked:
                    if self.duplicates.get(sample_hash) is not None:
                        logger.error(
                            "Possible backlog corruption for sample %d! "
                            "Please file a bug report. Trying to continue...",
                            sample.id)
                        continue

                    # submit one of the held-back samples as a new master
                    # analysis in case the analysis on the other instance
                    # failed and we have no result in the database yet. If all
                    # is well, this master should finish analysis very quickly
                    # using the stored result, causing all the duplicates to be
                    # submitted and finish quickly as well.
                    sample = sample_duplicates.pop()
                    self.duplicates[sample_hash] = {
                        'master': sample,
                        'duplicates': sample_duplicates,
                    }
                    submitted_cluster_duplicates.append(sample.id)
                    self.jobs.put(sample, True, self.queue_timeout)
                    del self.cluster_duplicates[sample_hash]

        if len(submitted_cluster_duplicates) > 0:
            logger.debug(
                "Submitted cluster duplicates (and potentially their "
                "duplicates) from backlog: %s", submitted_cluster_duplicates)

        return True

    def clear_stale_in_flight_samples(self):
        """ Clear any stale in-flight sample logs from the database. """
        try:
            cleared = self.db_con.clear_stale_in_flight_samples()
        except PeekabooDatabaseError as dberr:
            logger.error(dberr)
            cleared = False

        return cleared

    def submit_duplicates(self, sample_hash):
        """ Check if any samples have been held from processing as duplicates
        and submit them now. Clear the original sample whose duplicates have
        been submitted from the in-flight list.

        @param sample_hash: Hash of sample to check for duplicates
        """
        submitted_duplicates = []
        with self.duplock:
            # duplicates which have been submitted from the backlog still
            # report done but do not get registered as potentially having
            # duplicates because we expect the ruleset to identify them as
            # already known and process them quickly now that the first
            # instance has gone through full analysis. Therefore we can ignore
            # them here.
            if sample_hash not in self.duplicates:
                return

            # submit all samples which have accumulated in the backlog
            for sample in self.duplicates[sample_hash]['duplicates']:
                submitted_duplicates.append(sample.id)
                self.jobs.put(sample, True, self.queue_timeout)

            sample = self.duplicates[sample_hash]['master']
            try:
                self.db_con.clear_sample_in_flight(sample)
            except PeekabooDatabaseError as dberr:
                logger.error(dberr)

            del self.duplicates[sample_hash]

        logger.debug("%d: Cleared sample from in-flight list", sample.id)
        if len(submitted_duplicates) > 0:
            logger.debug(
                "Submitted duplicates from backlog: %s", submitted_duplicates)

    def done(self, sample):
        """ Perform cleanup actions after sample processing is done:
        1. Submit held duplicates and
        2. notify request handler thread that sample processing is done.

        @param sample: The Sample object to post-process. """
        self.submit_duplicates(sample.sha256sum)

    def dequeue(self):
        """ Remove a sample from the queue. Used by the workers to get their
        work. Blocks indefinitely until some work is available. If we want to
        wake the workers for some other reason, we send them a None item as
        ping. """
        return self.jobs.get(True)

    def shut_down(self):
        """ Trigger a shutdown of the queue including the workers. """
        logger.info("Queue shutdown requested. Signalling workers.")

        if self.ruleset_engine is not None:
            self.ruleset_engine.shut_down()

        if self.cluster_duplicate_handler is not None:
            self.cluster_duplicate_handler.shut_down()

        # tell all workers to shut down
        for worker in self.workers:
            worker.shut_down()

        # put a ping for each worker on the queue. Since they already all know
        # that they're supposed to shut down, each of them will only remove
        # one item from the queue and then exit, leaving the others for their
        # colleagues. For this reason this loop can't be folded into the above!
        for worker in self.workers:
            self.jobs.put(None)

    def close_down(self, timeout=None):
        """ Wait for workers to stop and free up resources. """
        if not timeout:
            timeout = self.shutdown_timeout

        logger.info("Closing down. Giving workers %d seconds to stop", timeout)

        # wait for workers to end
        interval = 1
        for attempt in range(1, timeout // interval + 1):
            still_running = []
            for worker in self.workers:
                if worker.is_alive():
                    still_running.append(worker)

            self.workers = still_running
            if len(self.workers) == 0:
                break

            time.sleep(interval)
            logger.debug('%d: %d workers still running', attempt,
                         len(self.workers))

        if len(self.workers) > 0:
            logger.error("Some workers refused to stop.")

        if self.cluster_duplicate_handler is not None:
            self.cluster_duplicate_handler.join()
        if self.ruleset_engine is not None:
            self.ruleset_engine.close_down()


class ClusterDuplicateHandler(threading.Thread):
    """ A housekeeping thread handling submission and cleanup cluster
    duplicates. """
    def __init__(self, job_queue, interval=5):
        self.shutdown_requested = threading.Event()
        self.shutdown_requested.clear()
        self.job_queue = job_queue
        self.interval = interval
        super().__init__(name="ClusterDuplicateHandler")

    def run(self):
        logger.debug("Cluster duplicate handler started.")

        while not self.shutdown_requested.wait(self.interval):
            logger.debug("Checking for samples in processing by other "
                         "instances to submit")
            # TODO: Error handling: How do we cause Peekaboo to exit with an
            # error from here? For now just keep trying and hope (database)
            # failure is transient.
            self.job_queue.clear_stale_in_flight_samples()
            self.job_queue.submit_cluster_duplicates()

        logger.debug("Cluster duplicate handler shut down.")

    def shut_down(self):
        """ Asynchronously initiate cluster duplicate handler shutdown. """
        self.shutdown_requested.set()


class Worker(threading.Thread):
    """ A Worker thread to process a sample. """
    def __init__(self, wid, job_queue, ruleset_engine, db_con):
        # whether we should run
        self.shutdown_requested = threading.Event()
        self.shutdown_requested.clear()
        self.worker_id = wid
        self.job_queue = job_queue
        self.ruleset_engine = ruleset_engine
        self.db_con = db_con
        super().__init__(name="Worker-%d" % wid)

    def run(self):
        while not self.shutdown_requested.is_set():
            logger.debug('Worker %d: Ready', self.worker_id)

            try:
                # wait blocking for next job (thread safe) with timeout
                sample = self.job_queue.dequeue()
            except queue.Empty:
                continue

            if sample is None:
                # we just got pinged
                continue

            logger.info('%d: Worker %d: Processing sample',
                        sample.id, self.worker_id)

            # The following used to be one big try/except block catching any
            # exception. This got complicated because in the case of
            # CuckooReportPending we use exceptions for control flow as well
            # (which might be questionable in itself). Instead of catching,
            # logging and ignoring errors here if workers start to die again
            # because of uncaught exceptions we should improve error handling
            # in the subroutines causing it.

            try:
                self.ruleset_engine.run(sample)
            except PeekabooAnalysisDeferred:
                logger.debug('%d: Report still pending', sample.id)
                continue

            if sample.result >= Result.failed:
                sample.dump_processing_info()

            sample.mark_done()

            logger.debug('%d: Saving results to database', sample.id)
            try:
                self.db_con.analysis_update(sample)
            except PeekabooDatabaseError as dberr:
                logger.error('%d: Failed to save analysis result to '
                             'database: %s', sample.id, dberr)
                # no showstopper, we can limp on without caching in DB

            self.job_queue.done(sample)

        logger.info('Worker %d: Stopped', self.worker_id)

    def shut_down(self):
        """ Asynchronously initiate worker shutdown. """
        self.shutdown_requested.set()



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# toolbox/                                                                    #
#         cortex.py                                                           #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2022 science + computing ag                              #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################

""" Interface to Cortex. """

import datetime
import http.cookiejar
import logging
import os
import threading
import enum

import cortex4py.api
import cortex4py.exceptions
import requests.sessions
import schema
import urllib3.util.retry

from peekaboo.exceptions import PeekabooException


logger = logging.getLogger(__name__)

class tlp(enum.Enum):
    WHITE = 0
    GREEN = 1
    AMBER = 2
    RED = 3

class CortexSubmitFailedException(PeekabooException):
    """ An exception raised if submitting a job to Cortex fails. """


class CortexAnalyzerReportMissingException(PeekabooException):
    """ An exception raised if an analysis still needs to be performed. """
    def __init__(self, analyzer):
        super().__init__("Cortex analysis report for analyzer '%s' is missing"
                         % analyzer)
        self.analyzer = analyzer


class CortexAnalyzerReport:
    """ Cortex analyzer report base class. """
    report_schema = schema.Schema(None, error='Subclass needs to provide a schema')

    report_schema_artifacts = [schema.Schema({
        "data": str,
        "dataType": str,
    # Possible future extensions
    #    "message": schema.Schema(schema.Or(str, None)),
    #    "tags": list,
    #    "tlp": int
    }, ignore_extra_keys=True)]

    def __init__(self, unvalidated_report):
        if unvalidated_report is None:
            unvalidated_report = {}

        # validate the report against subclass attribute report_schema
        self.report = self.report_schema.validate(unvalidated_report)

        self._domain_artifacts = self.get_filtered_elements_from_list_of_dicts(
                self.report.get('artifacts', []), 'dataType', 'domain', 'data', str)
        self._hash_artifacts = self.get_filtered_elements_from_list_of_dicts(
                self.report.get('artifacts', []), 'dataType', 'hash', 'data', str)
        self._ip_artifacts = self.get_filtered_elements_from_list_of_dicts(
                self.report.get('artifacts', []), 'dataType', 'ip', 'data', str)

    @classmethod
    def get_element_from_list_of_dicts(cls, list_, ident_key, ident_value, default={}):
        #pylint: disable=dangerous-default-value
        """ Search a list of dicts for an element with a matching value for an
            identifying key """
        element = [dictionary for dictionary in list_
            if dictionary.get(ident_key) == ident_value
        ]
        if len(element) != 1:
            return default
        return element[0]

    @classmethod
    def get_filtered_elements_from_list_of_dicts(
            cls, dictionary, ident_key, ident_value, filter_key, filter_data_type):
        """ Search a list of dicts for elements with a matching value for an
            identifying key and filter on filter_key with filter_data_type """
        elements = [artifact['data'] for artifact in dictionary
            if artifact.get(ident_key) == ident_value and
                filter_key in artifact and
                isinstance(artifact[filter_key], filter_data_type)
        ]
        return elements

    @property
    def domain_artifacts(self):
        """ Returns a list of domain artifacts. """
        return self._domain_artifacts

    @property
    def hash_artifacts(self):
        """ Returns a list of hash artifacts. """
        return self._hash_artifacts

    @property
    def ip_artifacts(self):
        """ Returns a list of ip artifacts. """
        return self._ip_artifacts


class CortexAnalyzer:
    """ Cortex analyzer base class. """
    name = 'unknown'
    reportclass = CortexAnalyzerReport


class CortexFileAnalyzer:
    """ An analyzer which accepts a file as main input. """
    @classmethod
    def get_submit_parameters(cls, sample, sample_tlp,
                              submit_original_filename=False):
        """ Return this analyzer's submit parameters for a given sample and a
        filtered, loggable version of them. """
        filename = sample.sha256sum

        # append file extension to aid backend analyzers in file type detection
        if sample.file_extension:
            filename = '%s.%s' % (filename, sample.file_extension)

        # override with the original file name if available
        if submit_original_filename and sample.filename:
            filename = sample.filename

        # log-friendly params prevent flooding the log with superfluous data
        # and can prevent information leakage
        log_params = {
            'data': '<%d bytes of sample data>' % sample.file_size,
            'dataProvided': True,
            'dataType': 'file',
            'tlp': sample_tlp.value,
            'parameters': {
                'filename': filename,
            }
        }

        params = log_params.copy()
        params['data'] = sample.content

        return params, log_params


class CortexHashAnalyzer(CortexAnalyzer):
    """ An analyzer which accepts hashes as main input. """
    @classmethod
    def get_submit_parameters(cls, sample, sample_tlp,
                              submit_original_filename=False):
        """ Return this analyzer's submit parameters for a given sample and a
        filtered, loggable version of them. """
        del submit_original_filename

        params = {
            'data': sample.sha256sum,
            'dataType': 'hash',
            'tlp': sample_tlp.value,
        }

        return params, params


class FileInfoAnalyzerReport(CortexAnalyzerReport):
    """ Represents a Cortex FileInfo_8_0 analysis JSON report. """

    report_schema = schema.Schema({
        "summary": {
            "taxonomies": [schema.Schema({
                "level": schema.Or("info", "malicious", "safe"),
                "namespace": "FileInfo",
            #    "predicate": str,
            #    "value": str
            }, ignore_extra_keys=True)]
        },
        "full": {
            "results": [
            {
                "submodule_name": "Basic properties",
                "results": [
                    {
                        "submodule_section_header": "Hashes",
                        "submodule_section_content": {
                            "md5": schema.Regex(r'^[0-9a-z]{32}$'),
                            "sha1": schema.Regex(r'^[0-9a-z]{40}$'),
                            "sha256": schema.Regex(r'^[0-9a-z]{64}$'),
                            "ssdeep": schema.Regex(r'^[0-9A-Za-z:+/]*$'),
                        }
                    },
                    {
                        # We consume further structures submodule_sections and
                        # explicitly check the submodule_section_header to not
                        # be "Hashes" or it will accept "Hashes"-structures with
                        # malfarmed hashes.
                        "submodule_section_header": schema.And(str, lambda s: s != "Hashes"),
                        "submodule_section_content": schema.Schema({
                            }, ignore_extra_keys=True)
                    },
                ],
                "summary": {
                    "taxonomies": [schema.Schema({
                        "level": schema.Or("info", "malicious", "safe"),
                        "namespace": "FileInfo",
                    #    "predicate": str,
                    #    "value": str
                    }, ignore_extra_keys=True)]
                }
            }
            ]
        },
        "success": bool,
        "artifacts": CortexAnalyzerReport.report_schema_artifacts,
        "operations": []
    })

    def __init__(self, unvalidated_report=None):
        """
        @param report: hash with report data from Cortex FileInfo Analyzer
        """
        super().__init__(unvalidated_report)

        basic_properties = self.get_element_from_list_of_dicts(
                self.report.get('full', []).get('results', {}),
                'submodule_name', 'Basic properties').get('results', [])
        self._hashes = self.get_element_from_list_of_dicts(
                basic_properties, 'submodule_section_header', 'Hashes').get(
                    'submodule_section_content', {})

    @property
    def sha256sum(self):
        """ Return the sha256 sum. """
        return self._hashes.get('sha256')

    @property
    def md5sum(self):
        """ Return the md5 sum. """
        return self._hashes.get('md5')

    @property
    def ssdeepsum(self):
        """ Return the ssdeep sum. """
        # TODO: think about if we want to compare ssdeep hashes
        return self._hashes.get('ssdeep')


class FileInfoAnalyzer(CortexFileAnalyzer):
    """ Interfaces with Cortex Analyzer FileInfo_8_0. """
    name = 'FileInfo_8_0'
    reportclass = FileInfoAnalyzerReport


class HybridAnalysisReport(CortexAnalyzerReport):
    """ Represents a Cortex HybridAnalysis_GetReport_1_0 analysis JSON
        report. """
    @property
    def full(self):
        """ Return the full report. """
        return self.report.get('full', None)


class HybridAnalysis(CortexFileAnalyzer):
    """ Interfaces with Cortex Analyzer HybridAnalysis_GetReport_1_0. """
    name = 'HybridAnalysis_GetReport_1_0'
    reportclass = HybridAnalysisReport


class VirusTotalQueryReport(CortexAnalyzerReport):
    """ Represents a Cortex VirusTotal_GetReport_3_0 analysis JSON report. """
    report_schema = schema.Schema({
        "summary": {
            "taxonomies": [
                {
                    "level": schema.Or("info", "malicious", "safe"),
                    "namespace": "VT",
                    "predicate": str,
                    "value": schema.Regex(r'^[0-9/]*$')
                }
            ]
        },
        "full": {
            "response_code": int,
            "resource": str,
            "verbose_msg": str
        },
        "success": bool,
        "artifacts": CortexAnalyzerReport.report_schema_artifacts,
        "operations": []
    })

    def __init__(self, unvalidated_report):
        super().__init__(unvalidated_report)

        self.taxonomies_vt = self.get_element_from_list_of_dicts(
                self.report.get('summary', {}).get('taxonomies'),
                'namespace', 'VT', {}
            )

    @property
    def n_of_all(self):
        """ n of all Virusscanners at VirusTotal have rated this file as
            malicious. """
        return int(self.taxonomies_vt.get('value', '-1/0').split('/')[0])

    @property
    def level(self):
        """ safe, suspicious, malicious """
        return self.taxonomies_vt.get('level', None)


class VirusTotalQuery(CortexHashAnalyzer):
    """ Interfaces with Cortex Analyzer VirusTotal_GetReport_3_0. """
    name = 'VirusTotal_GetReport_3_0'
    reportclass = VirusTotalQueryReport


class CuckooSandboxFileAnalysisReport(CortexAnalyzerReport):
    """ Represents a Cortex CuckooSandbox_File_Analysis_Inet_1_2 analysis JSON
        report. """
    def __init__(self, report):
        super().__init__(report)
        self.taxonomies = report.get("summary", {}).get("taxonomies", [{}])

    @property
    def signatures(self):
        """ Matched Cuckoo signatures. """
        return self.report.get('full', {}).get('Signatures', None)

    @property
    def malscore(self):
        """ Malscore n of 10 (might be bigger). """
        for tax in self.taxonomies:
            if tax.get('predicate') == 'Malscore':
                return float(tax['value'])
        return -1


class CuckooSandboxFileAnalysis(CortexFileAnalyzer):
    """ Interfaces with Cortex Analyzer CuckooSandbox_File_Analysis_Inet_1_2.
    """
    name = 'CuckooSandbox_File_Analysis_Inet_1_2'
    reportclass = CuckooSandboxFileAnalysisReport


class CAPEv2FileAnalysisReport(CortexAnalyzerReport):
    """ Represents a Cortex CAPESandbox_File_Analysis_Inet_0_1 analysis JSON
        report. """
    def __init__(self, report):
        super().__init__(report)
        self.taxonomies = report.get("summary", {}).get("taxonomies", [{}])

    @property
    def signatures(self):
        """ Matched CAPE signatures. """
        return self.report.get('full', {}).get('signatures', None)

    @property
    def malscore(self):
        """ Malscore n of 10 (might be bigger). """
        for tax in self.taxonomies:
            if tax.get('predicate') == 'Malscore':
                return float(tax['value'])
        return -1


class CAPEv2FileAnalysis(CortexFileAnalyzer):
    """ Interfaces with Cortex Analyzer CAPESandbox_File_Analysis_Inet_0_1. """
    name = 'CAPESandbox_File_Analysis_Inet_0_1'
    reportclass = CAPEv2FileAnalysisReport


class CortexReport:
    """ Meta class that either returns Cortex analysis reports or requests an
    analysis to be performed if it's not available yet. """
    def __init__(self):
        """ Initialize the report object. """
        self.reports = {}

    def register_report(self, analyzer, report):
        """ Register a report from an analyzer.

        @param analyzer: The analyzer the report is from.
        @type analyzer: CortexAnalyzer
        @param report: The report dict as returned by the API.
        @type report: dict
        """
        if analyzer.name in self.reports:
            logger.warning('Analysis report for %s already present - '
                           'replacing.', analyzer.name)

        self.reports[analyzer.name] = analyzer.reportclass(report)

    def get_report(self, analyzer):
        """ Try to retrieve an analyzer report from the sample.

        @param analyzer: Name of the analyzer
        @type analyzer: string
        @returns: CortexAnalyzerReport if present
        @raises: CortexAnalyzerReportMissingException if the report is not
                 yet present.
        """
        report = self.reports.get(analyzer.name)
        if report is None:
            raise CortexAnalyzerReportMissingException(analyzer)

        return report

    @property
    def FileInfoReport(self):
        """ Retrieve FileInfo analyzer report. """
        return self.get_report(FileInfoAnalyzer())

    @property
    def HybridAnalysisReport(self):
        """ Retrieve HybridAnalysis analyzer report. """
        return self.get_report(HybridAnalysis())

    @property
    def VirusTotalQueryReport(self):
        """ Retrieve VirusTotalQuery analyzer report. """
        return self.get_report(VirusTotalQuery())

    @property
    def CuckooSandboxFileReport(self):
        """ Retrieve CuckooSandboxFile analyzer report. """
        return self.get_report(CuckooSandboxFileAnalysis())

    @property
    def CAPEv2FileReport(self):
        """ Retrieve CAPEv2File analyzer report. """
        return self.get_report(CAPEv2FileAnalysis())


class CortexJob:
    """ Remember sample and submission time of a Cortex job. """
    def __init__(self, sample, analyzer):
        """ Initialise the CortexJob wrapper.

        @param sample: The sample in analysis.
        @type sample: Sample
        @param job: The Cortex analysis job
        @type job: cortex4py.controllers.JobController
        @param analyzer: Our analyzer object
        @type analyzer: CortexAnalyzer
        """
        self.__sample = sample
        self.__submission_time = datetime.datetime.utcnow()
        self.__analyzer = analyzer

    def is_older_than(self, seconds):
        """ Returns True if the difference between submission time and now,
        i.e. the age of the job, is larger than given number of seconds. """
        max_age = datetime.timedelta(seconds=seconds)
        return datetime.datetime.utcnow() - self.__submission_time > max_age

    @property
    def sample(self):
        """ Returns the sample the job is analyzing. """
        return self.__sample

    @property
    def analyzer(self):
        """ Returns the analyzer used to analyse the sample. """
        return self.__analyzer


class AllowedStatusRetry(urllib3.util.retry.Retry):
    """ A Retry class which has a list of allowed status codes, allowing to
    retry all status codes not explicitly allowed in a hard-core, catch-all
    manner. """
    def __init__(self, allowed_statuses=None, abort=None, **kwargs):
        super().__init__(**kwargs)
        self.allowed_statuses = allowed_statuses or set()
        # Event that is set if we're not to retry
        self.abort = abort

    def new(self, **kwargs):
        """ Adjusted shallow copy method to carry our parameters over into our
        copy. """
        if 'allowed_statuses' not in kwargs:
            kwargs['allowed_statuses'] = self.allowed_statuses
        if 'abort' not in kwargs:
            kwargs['abort'] = self.abort
        return super().new(**kwargs)

    def is_exhausted(self):
        """ Allow to abort a retry chain through an external signal. """
        if self.abort and self.abort.is_set():
            return True

        return super().is_exhausted()

    def is_retry(self, method, status_code, has_retry_after=False):
        """ Override Retry's is_retry to introduce our allowed statuses logic.
        """
        # we retry all methods so no check if method is retryable here

        if self.allowed_statuses and status_code not in self.allowed_statuses:
            return True

        return super().is_retry(method, status_code, has_retry_after)


class NocookiesPolicy(http.cookiejar.DefaultCookiePolicy):
    """ A cookie policy that denies to accept any cookies. """

    # CookiePolicy as a base class is not enough. CookieJar makes assumptions
    # about the expansive interface of DefaultCookiePolicy.

    def set_ok(self, cookie, request):
        """ No cookie will be accepted ever. """
        return False


class Cortex:
    """ Interfaces with a Cortex installation via its REST API. """
    def __init__(self, job_queue, url="http://localhost:9001", tlp=tlp.AMBER,
                 api_token="", poll_interval=5, submit_original_filename=True,
                 max_job_age=900, retries=5, backoff=0.5):
        """ Initialize the object.

        @param job_queue: The job queue to use from now on
        @type job_queue: JobQueue object
        @param url: Where to reach the Cortex REST API
        @type url: string
        @param tlp: colour according to traffic light protocol
        @type tlp: tlp
        @param api_token: API token to use for authentication
        @type api_token: string
        @param poll_interval: How long to wait inbetween job status checks
        @type poll_interval: int
        @param submit_original_filename: Whether to provide the original
                                         filename to Cortex to enhance
                                         analysis.
        @type submit_original_filename: bool
        @param max_job_age: How long to track jobs before declaring them
                            failed.
        @type max_job_age: int (seconds)
        @param retries: Number of retries on API requests
        @type retries: int
        @param backoff: Backoff factor for urllib3
        @type backoff: float
        """
        self.job_queue = job_queue
        self.shutdown_requested = threading.Event()
        self.shutdown_requested.clear()
        self.running_jobs = {}
        # reentrant because we're doing nested calls within critical sections
        self.running_jobs_lock = threading.RLock()
        self.url = url
        self.tlp = tlp
        self.api_token = api_token
        self.poll_interval = poll_interval
        self.submit_original_filename = submit_original_filename
        self.max_job_age = max_job_age

        # urrlib3 backoff formula:
        # <backoff factor> * (2 ^ (<retry count so far> - 1))
        # with second try intentionally having no sleep,
        # e.g. with retry count==5 and backoff factor==0.5:
        # try 1: fail, sleep(0.5*2^(1-1)==0.5*2^0==0.5*1==0.5->intentionally
        #   overridden to 0)
        # try 2: fail, sleep(0.5*2^(2-1)==0.5*2^1==1)
        # try 3: fail, sleep(0.5*2^(3-1)==0.5*2^2==2)
        # try 4: fail, sleep(0.5*2^(4-1)==0.5*2^3==4)
        # try 5: fail, abort, sleep would've been 8 before try 6
        #
        # Also, use allowed_methods=None to enable POST and other methods for
        # retry which aren't by default because they're not considered
        # idempotent. We assume that with the REST API a request either
        # succeeds or fails without residual effects, making them atomic and
        # idempotent.
        #
        # And finally we retry everything but a 200 response, which admittedly
        # is a bit hard-core but serves our purposes for now.
        retry_config = AllowedStatusRetry(
            total=retries, backoff_factor=backoff, allowed_methods=None,
            allowed_statuses=set([200]), abort=self.shutdown_requested)
        retry_adapter = requests.adapters.HTTPAdapter(max_retries=retry_config)
        self.session = requests.sessions.Session()
        self.session.mount('http://', retry_adapter)
        self.session.mount('https://', retry_adapter)
        # attach a cookie policy that refuses to learn any cookies from
        # responses. This is because Cortex sometimes hands out CSRF tokens and
        # even session cookies in response to our bearer-token-authenticated
        # API requests which we don't need but have the potential to confuse
        # Cortex on subsequent requests.
        self.session.cookies = requests.cookies.RequestsCookieJar(
            NocookiesPolicy())

        self.api = None
        self.tracker = None

    def register_running_job(self, job_id, job):
        """ Register a job as running. Detect if another sample has already
        been registered with the same job ID which obviously must never happen
        because it corrupts our internal housekeeping. Guarded by a lock
        because multiple worker threads will call this routine and check for
        collision and update of job log might otherwise race each other.

        @param job_id: Cortex job ID of the job to register as running.
        @type job_id: string
        @param job: Our Cortex job wrapper object
        @type job: CortexJob

        @returns: None
        @raises: CortexSubmitFailedException on job id collision """
        with self.running_jobs_lock:
            if (job_id in self.running_jobs and
                    self.running_jobs[job_id] is not job):
                logger.warning(
                    '%d: A job with ID %s already registered as running '
                    'for different sample %d will be marked failed',
                    job.sample.id, job_id,
                    self.running_jobs[job_id].sample.id)
                self.resubmit_as_failed(job_id)

            self.running_jobs[job_id] = job

    def deregister_running_job(self, job_id):
        """ Deregister a running job by job id.

        @param job_id: Cortex job ID of the job do deregister.
        @type job_id: string
        @returns: Sample object of the job or None if job not found. """
        with self.running_jobs_lock:
            return self.running_jobs.pop(job_id, None)

    def deregister_running_job_if_too_old(self, job_id, max_age):
        """ Check if a job has gotten too old and remove it from the list of
        running jobs if so.

        @param job_id: Cortex job ID of the job do deregister.
        @type job_id: string
        @returns: Sample object of the job or None if job not found. """
        with self.running_jobs_lock:
            if self.running_jobs[job_id].is_older_than(max_age):
                return self.deregister_running_job(job_id)

        return None

    def resubmit_with_analyzer_report(self, job_id):
        """ Resubmit a sample to the job queue after an analyzer report became
        available. Retrieves the report from Cortex.

        @param job_id: ID of job which has finished.
        @type job_id: string
        @param report: Analyzer report
        @type report: dict

        @returns: None
        """
        logger.debug("Analysis done for job %s", job_id)

        job = self.deregister_running_job(job_id)
        if job is None:
            logger.debug('No job found for job ID %s', job_id)
            return None

        logger.debug('%d: Requesting Cortex report', job.sample.id)
        try:
            report = self.api.jobs.get_report(job_id)
            # register this job's analysis report with our main report object
            job.sample.cortex_report.register_report(
                job.analyzer, report.report)
        except cortex4py.exceptions.CortexException as error:
            logger.error('Retrieval of report from Cortex failed: %s', error)
            # mark analysis as failed if we could not get the report e.g.
            # because it was corrupted or the API connection failed.
            job.sample.mark_cortex_failure()
        except schema.SchemaError as error:
            logger.warning('Report returned from Cortex contained '
                           'invalid data: %s', error)
            job.sample.mark_cortex_failure()

        self.job_queue.submit(job.sample)
        return None

    def resubmit_as_failed(self, job_id):
        """ Resubmit a sample to the job queue with a failure report if the
        Cortex job has failed for some reason.

        @param job_id: ID of job that failed.
        @type job_id: string
        """
        job = self.deregister_running_job(job_id)
        if job is not None:
            job.sample.mark_cortex_failure()
            self.job_queue.submit(job.sample)

    def resubmit_as_failed_if_too_old(self, job_id, max_age):
        """ Resubmit a sample to the job queue with a failure report if the
        Cortex job has been running for too long.

        @param job_id: ID of job to check.
        @type job_id: string
        @param max_age: maximum job age in seconds
        @type max_age: int
        """
        job = self.deregister_running_job_if_too_old(job_id, max_age)
        if job is not None:
            logger.warning("Dropped job %s because it has been running for "
                           "too long", job_id)
            job.sample.mark_cortex_failure()
            self.job_queue.submit(job.sample)

    def submit(self, sample, analyzer):
        """ Submit a sample to Cortex for analysis.
        @param sample: Sample to submit.
        @type sample: Sample

        @raises: CortexSubmitFailedException if submission failed
        @returns: ID of the submitted Cortex job.
        """
        params, log_params = analyzer.get_submit_parameters(
            sample, self.tlp, self.submit_original_filename)

        logger.debug("Creating Cortex job with analyzer %s and "
                     "parameters %s", analyzer.name, log_params)
        try:
            job = self.api.analyzers.run_by_name(analyzer.name, params)
        except cortex4py.exceptions.CortexException as error:
            raise CortexSubmitFailedException(
                'Error submitting Cortex job: %s' % error)

        self.register_running_job(job.id, CortexJob(sample, analyzer))
        return job.id

    def start_tracker(self):
        """ Start tracking running jobs in a separate thread. """
        # do a simple initial test of connectivity. With this we require the
        # API to be reachable at startup (with the usual retries to account for
        # a bit of a race condition in parallel startup) but later on hope
        # that all errors are transient and retry endlessly
        self.api = cortex4py.api.Api(
            self.url, self.api_token, session=self.session)
        try:
            self.api.analyzers.find_all({}, range='all')
        except cortex4py.exceptions.CortexException as error:
            logger.error('Initial connection to Cortex API failed: %s', error)
            return False

        self.tracker = threading.Thread(target=self.track,
                                        name="CortexJobTracker")
        self.tracker.start()
        return True

    def track(self):
        """ Do the polling for finished jobs. """
        while not self.shutdown_requested.wait(self.poll_interval):
            # no lock, atomic, copy() because keys() returns an iterable view
            # instead of a fresh new list in python3
            running_jobs = self.running_jobs.copy().keys()

            # somewhat inefficient for the potential number of requests per
            # poll round but more efficient in that it does not download an
            # ever growing list of jobs using tasks/list. Might also take
            # quite a while to get to all jobs if retries happen on each
            # request.
            # A call to get data about multiple tasks in one go would be nice
            # here. tasks/list could be used with the minimum job number as
            # offset and spread between highest and lowest job id as limit *if*
            # its output was sorted by job ID. Apparently limit and offset are
            # only meant to iterate over the job list in blocks but not to
            # return data about a specific range of job IDs from that list.
            for job_id in running_jobs:
                cortexjob = None
                try:
                    cortexjob = self.api.jobs.get_by_id(job_id)
                except cortex4py.exceptions.CortexException as error:
                    logger.error('Querying Cortex job status failed: %s', error)
                    # ignore and retry on next polling run
                    continue

                if cortexjob.status in ['Success']:
                    self.resubmit_with_analyzer_report(job_id)
                    continue

                if cortexjob.status in ['Failure', 'Deleted']:
                    logger.warning("Dropping job %s because it has failed "
                                   "in Cortex", job_id)
                    self.resubmit_as_failed(job_id)
                    continue

                # drop jobs which have been running for too long. This is
                # mainly to prevent accumulation of jobs in our job list which
                # will never finish. We still want to wait for jobs to finish
                # even though our client might not be interested any more so
                # that we have the result cached for the next time we get the
                # same sample.
                self.resubmit_as_failed_if_too_old(job_id, self.max_job_age)

        logger.debug("Cortex job tracker shut down.")

    def shut_down(self):
        """ Request the module to shut down, used by the signal handler. """
        logger.debug("Cortex job tracker shutdown requested.")
        self.shutdown_requested.set()

    def close_down(self):
        """ Close down tracker resources, particularly, wait for the thread to
        terminate. """
        if self.tracker:
            self.tracker.join()
            self.tracker = None



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# sample.py                                                                   #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2022 science + computing ag                              #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


import datetime
import enum
import os
import hashlib
import json
import random
import re
import shutil
import string
import logging
import tempfile
# python 3's open with encoding parameter and implicit usage of the system
# locale-specified encoding
from builtins import open
from peekaboo.ruleset import Result


logger = logging.getLogger(__name__)


class SampleFactory:
    """ A class for churning out loads of mostly identical sample objects.
    Contains all the global configuration data and object references each
    sample needs and thus serves as a registry of potential API breakage
    perhaps deserving looking into. """
    def __init__(self, processing_info_dir):
        # configuration
        self.processing_info_dir = processing_info_dir

    def make_sample(self, content, name=None, content_type=None,
                    content_disposition=None):
        """ Create a new Sample object based on the factory's configured
        defaults and variable parameters. """
        return Sample(content, name, content_type, content_disposition,
                      self.processing_info_dir)


@enum.unique
class JobState(enum.Enum):
    """ Enumeration of states a job processing some sample can be in. """
    ACCEPTED = 1
    FINISHED = 2


class Sample:
    """
    This class handles and describes samples to be analysed by Peekaboo.

    A sample has properties like filename, MIME type, checksum or file size.
    These are accessible as properties. Most properties determine their value
    on first access, especially if that determination is somewhat expensive
    such as the file checksum.
    """
    def __init__(self, content, filename=None, content_type=None,
                 content_disposition=None,
                 processing_info_dir=None, job_id=None):
        # we do neither need nor accept for path traversal attack avoidance
        # full paths
        if filename is not None:
            filename = os.path.basename(filename)

        self.__content = content
        self.__filename = filename
        self.__content_type = content_type
        self.__content_disposition = content_disposition
        self.__cuckoo_failed = False
        self.__cuckoo_report = None
        self.__cortex_failed = False
        self.__cortex_report = None
        self.__oletools_report = None
        self.__filetools_report = None
        self.__knowntools_report = None
        self.__id = job_id
        self.__state = JobState.ACCEPTED
        self.__result = Result.unchecked
        self.__reason = None
        self.__report = []
        self.__sha256sum = None
        self.__file_extension = None
        self.__processing_info_dir = processing_info_dir

    @property
    def filename(self):
        """ Returns the name of the sample file, i.e. the basename without path
        but including the file extension. """
        return self.__filename

    @property
    def content(self):
        """ Returns the content of the sample file. """
        return self.__content

    @property
    def result(self):
        """ Returns the overall evaluation result for this sample.

        @returns: peekaboo.ruleset.Result """
        return self.__result

    @property
    def reason(self):
        """ Gets the reason given by the rule determining the result which
        ended up as the overall evaluation result of this sample.

        @returns: string """
        return self.__reason

    @property
    def peekaboo_report(self):
        """ Return Peekaboo's report meant for the client, detailing what's
        been found on this sample.

        @returns: List of strings.
        """
        # This message used to be:
        # "Die Datei \"%s\" wurde als \"%s\" eingestuft\n\n"
        # Changed intentionally to not trigger configured god/bad matching
        # patterns in clients (e.g. AMaViS) any more since we switched to
        # reporting an overall analysis batch result.
        return self.__report + [_("File \"%s\" is considered \"%s\"")
                                % (self.__filename, self.__result.name)]

    @property
    def id(self):
        """ Return the (job) ID of this sample. """
        return self.__id

    def update_id(self, job_id):
        """ Update the (job) ID of this sample. """
        self.__id = job_id

    @property
    def state(self):
        """ Tells whether the analysis of the sample is done, i.e. a final
        verdict has been reached and a result and reason are available. """
        return self.__state

    def mark_done(self):
        """ Mark this sample as done, i.e. fully analysed and verdict reached.
        """
        self.__state = JobState.FINISHED

    def add_rule_result(self, res):
        """ Add a rule result to the sample. This also adds a message about
        this to the report and updates the overall analysis result (so far).
        """
        res_str = "%s" % res
        logger.debug("%d: Adding rule result %s", self.__id, res_str)
        self.__report.append(res_str)

        logger.debug("%d: Current overall result: %s, new rule result: %s",
                     self.__id, self.__result, res.result)
        # check if result of this rule is worse than what we know so far
        if res.result >= self.__result:
            self.__result = res.result
            self.__reason = res.reason

    def dump_processing_info(self):
        """
        Saves the Cuckoo report as HTML + JSON
        to a directory named after the job hash.
        """
        if not self.__processing_info_dir:
            logger.debug('Not dumping processing info because no path for the '
                         'data is unconfigured.')
            return

        now = datetime.datetime.now().strftime("%Y%m%dT%H%M%S")
        dump_dir = os.path.join(
            self.__processing_info_dir, "%s-%s" % (now, self.sha256sum))
        if not os.path.isdir(dump_dir):
            try:
                os.makedirs(dump_dir, 0o770)
            except OSError as oserr:
                logger.error('Failed to create dump directory %s: %s',
                             dump_dir, oserr)
                return

        logger.debug('%d: Dumping processing info to %s',
                     self.__id, dump_dir)

        # Peekaboo's report
        peekaboo_report = os.path.join(dump_dir, 'report.txt')
        try:
            with open(peekaboo_report, 'w+') as pr_file:
                pr_file.write('Declared file name: %s\n' % self.__filename)
                pr_file.write(
                    'Declared content type: %s\n' % self.__content_type)
                pr_file.write(
                    'Declared content disposition: %s\n' %
                    self.__content_disposition)
                if self.__report:
                    pr_file.write('\n'.join(self.__report + [""]))
        except (OSError, IOError) as error:
            logger.error('Failure to write report file %s: %s',
                         peekaboo_report, error)
            return

        # store malicious sample along with the reports
        if self.__result == Result.bad:
            sample_dump = os.path.join(dump_dir, 'sample.bin')
            try:
                with open(sample_dump, 'wb') as dump_file:
                    dump_file.write(self.__content)
            except (shutil.Error, IOError, OSError) as error:
                logger.error('Failure to dump sample file to dump '
                             'directory: %s', error)
                return

        # Cuckoo report
        if self.__cuckoo_report:
            cuckoo_report = os.path.join(dump_dir, 'cuckoo_report.json')
            try:
                with open(cuckoo_report, 'wb+') as cr_json_file:
                    cr_json = json.dumps(self.__cuckoo_report.dump,
                                         indent=1, ensure_ascii=True)
                    cr_json_file.write(cr_json.encode('ascii'))
            except (OSError, IOError) as error:
                logger.error('Failure to dump json report to %s: %s',
                             cuckoo_report, error)
                return

    @property
    def sha256sum(self):
        """ Returns the SHA256 checksum/fingerprint of this sample. Determines
        it automatically on first call. """
        if not self.__sha256sum:
            self.__sha256sum = hashlib.sha256(self.__content).hexdigest()

        return self.__sha256sum

    @property
    def content_disposition(self):
        """ Returns the content disposition in the original email, e.g. inline
        or attachment, None if not available. """
        return self.__content_disposition

    # for compat with ruleset
    @property
    def name_declared(self):
        """ Returns the name declared by the sample as its original filename,
        None if not available. """
        return self.__filename

    @property
    def file_extension(self):
        """ Determines the file extension of this sample. """
        if self.__file_extension:
            return self.__file_extension

        # no extension if we have not filename to extract it from
        if self.__filename is None:
            return None

        ext = os.path.splitext(self.__filename)[1][1:]

        # sanity checks, where splitext already promises:
        # - ext is empty or
        # - begins with a period and contains at most one period.
        #   [and we already strippped that period]
        # - Leading periods on the basename are ignored;
        # We extend that to:
        # - allow only ascii (no Unicode)
        # - allow only printable characters (no control codes)
        # - allow only alphanumeric characters and some symbols (arbitrarily
        #   chosen based on file extension list in Wikipedia and personal
        #   experience)
        allowed_characters = string.ascii_letters + string.digits

        # we're seeing attachments whose declared filenames include what seems
        # to be meant as query strings (e.g. foo.jpg?resize=600,510). Since
        # there are no file extensions we know of containing those characters,
        # we do not allow them. Since we cannot find any document stating that
        # the name parameter of header Content-Type can contain URLs/URIs,
        # we're not even attempting to go down the rabbit hole of parsing it as
        # such to avoid the inevitable fallout from it. Instead we rather not
        # extract any extension at all.
        allowed_characters += string.punctuation.translate(
                str.maketrans('', '', '?;&'))

        # test works indirectly by stripping what is allowed from beginning and
        # end of the extension and checking that nothing remains.
        if ext.strip(allowed_characters):
            return None

        self.__file_extension = ext
        return self.__file_extension

    @property
    def type_declared(self):
        """ Returns the MIME type declared by the original MIME part, None if
        not available. """
        return self.__content_type

    @property
    def file_size(self):
        """ Return sample file size based on content length. """
        return len(self.__content)

    @property
    def cuckoo_failed(self):
        """ Returns whether the Cuckoo analysis failed. """
        return self.__cuckoo_failed

    @property
    def cuckoo_report(self):
        """ Returns the cuckoo report """
        return self.__cuckoo_report

    @property
    def cortex_failed(self):
        """ Returns whether a Cortex analysis failed. """
        return self.__cortex_failed

    @property
    def cortex_report(self):
        """ Returns the Cortex report. """
        return self.__cortex_report

    @property
    def oletools_report(self):
        """ Returns the oletools report """
        return self.__oletools_report

    @property
    def filetools_report(self):
        """ Returns the filetools report """
        return self.__filetools_report

    @property
    def knowntools_report(self):
        """ Returns the knowntools report """
        return self.__knowntools_report

    def mark_cuckoo_failure(self):
        """ Records whether Cuckoo analysis failed. """
        self.__cuckoo_failed = True

    def register_cuckoo_report(self, report):
        """ Records a Cuckoo report for later evaluation. """
        self.__cuckoo_report = report

    def mark_cortex_failure(self):
        """ Records whether Cortex analysis failed. """
        self.__cortex_failed = True

    def register_cortex_report(self, report):
        """ Records a Cortex report for later evaluation. """
        self.__cortex_report = report

    def register_oletools_report(self, report):
        """ Records a Oletools report for alter evaluation. """
        self.__oletools_report = report

    def register_filetools_report(self, report):
        """ Records a Filetools report for alter evaluation. """
        self.__filetools_report = report

    def register_knowntools_report(self, report):
        """ Records a Knowntools report for alter evaluation. """
        self.__knowntools_report = report

    def __str__(self):
        """ A string representation with selected information for positive
        identification. ID should be enough but may be uninitialised. Should
        not modify object state, i.e. should not use properties. """
        return f"<Sample(id='{self.__id}')>"

    __repr__ = __str__



#!/usr/bin/env python
# encoding: utf-8

###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# peekaboo-util.py                                                            #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2022 science + computing ag                              #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################

""" The peekaboo-util command line interface. """

import argparse
import json
import logging
import os
import re
import socket
import sys
import time
import urllib.parse
import requests


logging.basicConfig()
logger = logging.getLogger(__name__)


class PeekabooUtil:
    """ Utility fo interface with Peekaboo API over the socket connection """
    def __init__(self, url, polling_interval, timeout):
        logger.debug('Initialising PeekabooUtil')
        logger.debug('Opening connection %s', url)
        self.url = url
        self.polling_interval = polling_interval
        self.timeout = timeout

    def ping(self, _):
        """ Send ping request to daemon and optionally print response. """
        logger.debug("Sending ping...")
        try:
            r = requests.get(
                urllib.parse.urljoin(self.url, "/ping"),
                timeout=self.timeout)
            pong = r.json()
        except socket.error as error:
            logger.error("Error communicating with daemon: %s", error)
            return 2
        except json.decoder.JSONDecodeError as error:
            logger.error("Error decoding the response: %s", error)
            return 2

        if not isinstance(pong, dict):
            logger.error("Invalid response from daemon: %s", pong)
            return 2

        reqtype = r.request
        response = pong.get('answer')
        if reqtype is None or response is None:
            logger.error("Incomplete response from daemon: %s", pong)
            return 2

        if reqtype != 'ping' and response != 'pong':
            logger.error("Response is not a pong")
            return 2

        logger.info('Pong received.')
        return 0

    def scan_file(self, args):
        """ Scan the supplied filenames with peekaboo and output result """
        jobs = []
        for filename in args.filename:
            logger.debug('Submitting file %s', filename)

            content_type = None
            if args.content_type:
                ct_popped = args.content_type.pop(0)
                # empty string is treated as no selection to allow specificaion
                # for later file arguments still
                if ct_popped:
                    content_type = ct_popped
                    logger.debug('Using content type %s', content_type)

            content_disposition = None
            if args.content_disposition:
                cd_popped = args.content_disposition.pop(0)
                if cd_popped:
                    content_disposition = cd_popped
                    logger.debug(
                        'Using content disposition %s', content_disposition)

            with open(filename, 'rb') as upload_file:
                submit_name = os.path.basename(filename)
                files = {'file': (submit_name, upload_file, content_type)}
                headers = {'x-content-disposition': content_disposition}

                # NOTE: As of this writing, requests via urllib3 1.26.x encodes
                # filenames according to a WHATWG HTML5 spec draft using
                # percent encoding and backslash escaping. Peekaboo via sanic
                # only supports part of the current WHATWG HTML5 spec (no
                # longer draft) by reverting double quote escaping (%22 -> ").
                # We cannot (without serious hacking) switch requests to use
                # RFC2231 encoding from here. Therefore filenames containing
                # special characters can not currently be correctly transferred
                # using this command. urllib3 has already been adjusted to
                # default to the current version of the HTML5 escaping scheme.
                # So sanic hopefully doing the same there's a chance this will
                # sort itself out by itself eventually.
                response = requests.post(urllib.parse.urljoin(
                    self.url, '/v1/scan'), files=files, headers=headers,
                    timeout=self.timeout)

            json_resp = response.json()
            job_id = json_resp.get('job_id')
            if job_id is None:
                logger.error('Job submission failed: %s', json_resp.get(
                    'message', 'No explanation given by Peekaboo'))
                continue

            logger.debug('File %s submitted as job %d', filename, job_id)

            jobs.append(job_id)

        exit_code = 0
        while True:
            jobs_left = []
            for job in jobs:
                response = requests.get(urllib.parse.urljoin(
                    self.url, f'/v1/report/{job}'), timeout=self.timeout)

                if response.status_code == 404:
                    jobs_left.append(job)
                    continue

                json_resp = response.json()
                if json_resp['result'] == 'bad':
                    print(f"{job}: The file has been categorized 'bad'")
                    exit_code = 1

                logger.info("%d: %s", job, json_resp['reason'])

            if not jobs_left:
                break

            jobs = jobs_left
            time.sleep(self.polling_interval)

        return exit_code

def main():
    """ The peekaboo-util main program. """
    # do not use store_true and store_false because it initializes the dest to
    # its opposite instead of None which interferes with our merging-down of
    # global parser values if not specified per command
    common_args = [
        {
            'args': ['-v', '--verbose'],
            'kwargs': {
                'action': 'store_const', 'const': True, 'required': False,
                'help': 'List results of all files not only bad ones'},
        }, {
            'args': ['-vv', '--verbose2'],
            'kwargs': {
                'action': 'store_const', 'const': True, 'required': False,
                'help': 'List detailed analysis results of every rule'},
        }, {
            'args': ['-d', '--debug'],
            'kwargs': {
                'action': 'store_const', 'const': True, 'required': False,
                'help': 'Output additional diagnostics'},
        }, {
            'args': ['-t', '--timeout'],
            'kwargs': {
                'type': float, 'required': False, 'default': None,
                'help': 'Communications timeout'},
        }, {
            'args': ['-i', '--polling-interval'],
            'kwargs': {
                'type': int, 'required': False, 'default': 5,
                'help': 'Polling interval'},
        }, {
            'args': ['-r', '--remote-url'],
            'kwargs': {
                'action': 'store', 'required': False,
                'default': 'http://127.0.0.1:8100/v1/',
                'help': 'URL to Peekaboo instance e.g. '
                        'http://127.0.0.1:8100/v1/'},
        }
    ]

    # using the common parser as parent to the command parser does not work as
    # the subcommand arguments will completely replace the command parser
    # defaults and values even if not specified on the command line
    common_parser = argparse.ArgumentParser(add_help=False)
    command_parser = argparse.ArgumentParser()
    for common_arg in common_args:
        args, kwargs = common_arg['args'], common_arg['kwargs']
        global_kwargs = kwargs.copy()

        # subcommands args must not have defaults for the global args and
        # potentially their defaults to be queried
        kwargs.pop('default', None)
        common_parser.add_argument(*args, **kwargs)

        # again, dest can not be the same because the subparser arguments will
        # completely replace the command parser value and defaults even if not
        # specified on the command line
        dest = kwargs.get('dest')
        if dest is None:
            dest = args[1].lstrip('-').replace('-', '_')
        common_arg['dest'] = dest
        global_kwargs['dest'] = "global_%s" % dest
        command_parser.add_argument(*args, **global_kwargs)

    command_parser.set_defaults(func=PeekabooUtil.ping)

    subparsers = command_parser.add_subparsers(help='commands')
    scan_file_parser = subparsers.add_parser(
        'scan-file', help='Scan a file and report it', parents=[common_parser])
    scan_file_parser.add_argument('-f', '--filename', action='append', required=True,
                                  help='Path to the file to scan. Can be given more '
                                       'than once to scan multiple files.')
    scan_file_parser.add_argument(
        '-c', '--content-type', action='append', required=False,
        help='Content type of file to scan. Can be given more than once and '
        'has to match the file list in order.')
    scan_file_parser.add_argument(
        '-o', '--content-disposition', action='append', required=False,
        help='Content disposition of file to scan. Can be given more than '
        'once and has to match the file list in order.')

    scan_file_parser.set_defaults(func=PeekabooUtil.scan_file)

    ping_parser = subparsers.add_parser(
        'ping', help='Ping the daemon', parents=[common_parser])
    ping_parser.set_defaults(func=PeekabooUtil.ping)

    args = command_parser.parse_args()

    # merge down global values or defaults
    for common_arg in common_args:
        dest = common_arg['dest']

        # dest is not present if no subcommand was specified, so check with
        # hasattr as well
        if not hasattr(args, dest) or getattr(args, dest) is None:
            global_value = getattr(args, "global_%s" % dest)
            setattr(args, dest, global_value)

    logger.setLevel(logging.ERROR)
    if args.verbose:
        logger.setLevel(logging.INFO)
    if args.verbose2 or args.debug:
        logger.setLevel(logging.DEBUG)

    try:
        util = PeekabooUtil(
            args.remote_url, args.polling_interval, args.timeout)
    except socket.error as error:
        logger.error("Error connecting to peekaboo: %s", error)
        return 2

    return args.func(util, args)


if __name__ == "__main__":
    sys.exit(main())



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# server.py                                                                   #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2022 science + computing ag                              #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################

""" This module implements the Peekaboo server, i.e. the frontend to the
client. """

import asyncio
import email.utils
import logging
import urllib.parse

import sanic
import sanic.headers
import sanic.response

from peekaboo.db import PeekabooDatabaseError

logger = logging.getLogger(__name__)

class PeekabooServer:
    """ A class wrapping the server components of Peekaboo. """
    def __init__(self, host, port, job_queue, sample_factory,
                 request_queue_size, db_con):
        """ Initialise a new server and start it. All error conditions are
        returned as exceptions.

        @param host: The local address to bind the socket to.
        @type host: String
        @param port: The local port to listen on for client connections.
        @type port: int
        @param job_queue: A reference to the job queue for submission of
                          samples.
        @type job_queue: JobQueue
        @param sample_factory: A reference to a sample factory for creating new
                               samples.
        @type sample_factory: SampleFactory
        @param request_queue_size: Number of requests that may be pending on
                                   the socket.
        @type request_queue_size: int
        """
        logger.debug('Starting up server.')
        self.app = sanic.Sanic("PeekabooAV", configure_logging=False)
        self.app.config.FALLBACK_ERROR_FORMAT = "json"

        # silence sanic to a reasonable amount
        logging.getLogger('sanic.root').setLevel(logging.WARNING)
        logging.getLogger('sanic.access').setLevel(logging.WARNING)

        self.loop = asyncio.get_event_loop()
        self.server_coroutine = self.app.create_server(
            host=host, port=port, return_asyncio_server=True,
            backlog=request_queue_size,
            asyncio_server_kwargs=dict(start_serving=False))
        self.server = None
        self.job_queue = job_queue
        self.sample_factory = sample_factory
        self.db_con = db_con
        # remember for diagnostics
        self.host = host
        self.port = port

        self.app.add_route(self.hello, '/')
        self.app.add_route(self.ping, '/ping')
        self.app.add_route(self.scan, "/v1/scan", methods=['POST'])
        self.app.add_route(
            self.report, '/v1/report/<job_id:int>', methods=['GET'])

    async def hello(self, _):
        """ hello endpoint as fallback and catch all

        @returns: hello world json response
        """
        return sanic.response.json({'hello': 'PeekabooAV'})

    async def ping(self, _):
        """ ping endpoint for diagnostics

        @returns: pong json response
        """
        return sanic.response.json({'answer': 'pong'})

    async def scan(self, request):
        """ scan endpoint for job submission

        @param request: sanic request object
        @type request: sanic.Request
        @returns: json response containing ID of newly created job
        """
        # this is sanic's multipart/form-data parser in a version that knows
        # that our file field contains binary data. This allows transferring
        # files without a filename. The generic parser would treat those as
        # text fields and try to decode them using the form charset or UTF-8 as
        # a fallback and cause errors such as: UnicodeDecodeError: 'utf-8'
        # codec can't decode byte 0xc0 in position 1: invalid start byte
        content_type, parameters = sanic.headers.parse_content_header(
            request.content_type)

        # application/x-www-form-urlencoded is inefficient at transporting
        # binary data. Also it needs a separate field to transfer the filename.
        # Make clear here that we do not support that format (yet).
        if content_type != 'multipart/form-data':
            logger.error('Invalid content type %s', content_type)
            return sanic.response.json(
                {'message': 'Invalid content type, use multipart/form-data'},
                400)

        boundary = parameters["boundary"].encode("utf-8")
        form_parts = request.body.split(boundary)
        # split above leaves preamble in form_parts[0] and epilogue in
        # form_parts[2]
        num_fields = len(form_parts) - 2
        if num_fields <= 0:
            logger.error('Invalid MIME structure in request, no fields '
                         'or preamble or epilogue missing')
            return sanic.response.json(
                {'message': 'Invalid MIME structure in request'}, 400)

        if num_fields != 1:
            logger.error('Invalid number of fields in form: %d', num_fields)
            return sanic.response.json(
                {'message': 'Invalid number of fields in form, we accept '
                    'only one field "file"'}, 400)

        form_part = form_parts[1]
        file_name = None
        content_type = None
        field_name = None
        line_index = 2
        line_end_index = 0
        while line_end_index != -1:
            line_end_index = form_part.find(b'\r\n', line_index)
            # this constitutes a hard requirement for the multipart headers
            # (and filenames therein) to be UTF-8-encoded. There are some
            # obscure provisions for transferring an encoding in RFC7578
            # section 5.1.2 for HTML forms which don't apply here so its
            # fallback to UTF-8 applies. This is no problem for our field name
            # (ASCII) and file names in RFC2231 encoding. For HTML5-style
            # percent-encoded filenames it means that whatever isn't
            # percent-encoded needs to be UTF-8 encoded. There are no rules in
            # HTML5 currently to percent-encode any UTF-8 byte sequences.
            form_line = form_part[line_index:line_end_index].decode('utf-8')
            line_index = line_end_index + 2

            if not form_line:
                break

            colon_index = form_line.index(':')
            idx = colon_index + 2
            form_header_field = form_line[0:colon_index].lower()

            # parse_content_header() reverts some of the percent encoding as
            # per HTML5 WHATWG spec. As it is a "living standard" (i.e. moving
            # target), it has changed over the years. There used to be
            # backslash doubling and explicit control sequence encoding. As of
            # this writing this has been changed to escaping only newline,
            # linefeed and double quote. Sanic only supports the double quote
            # part of that: %22 are reverted back to %. Luckily this interacts
            # reasonably well with RFC2231 decoding below since that would do
            # the same.
            #
            # There is no way to tell what version of the standard (or draft
            # thereof) the client was following when encoding. It seems accepted
            # practice in the browser world to just require current versions of
            # everything so their behaviour hopefully converges eventually.
            # This is also the reason why we do not try to improve upon it here
            # because it's bound to become outdated.
            #
            # NOTE: Since we fork the sanic code here we need to keep track of
            # its changes, particularly how it interacts with RFC2231 encoding
            # if escaping of the escape character %25 is ever added to the
            # HTML5 WHATWG spec. In that case parse_content_header() would
            # start breaking the RFC2231 encoding which would explain why its
            # use is forbidden in RFC7578 section 4.2 via RFC5987.
            form_header_value, form_parameters = sanic.headers.parse_content_header(
                form_line[idx:]
            )

            if form_header_field == 'content-disposition':
                field_name = form_parameters.get('name')
                file_name = form_parameters.get('filename')

                # non-ASCII filenames in RFC2231, "filename*" format
                if file_name is None and form_parameters.get('filename*'):
                    encoding, _, value = email.utils.decode_rfc2231(
                        form_parameters['filename*']
                    )
                    file_name = urllib.parse.unquote(value, encoding=encoding)
            elif form_header_field == 'content-type':
                content_type = form_header_value

        if field_name != 'file':
            logger.error('Field file missing from request')
            return sanic.response.json(
                {'message': 'Field "file" missing from request'}, 400)

        file_content = form_part[line_index:-4]
        content_disposition = request.headers.get('x-content-disposition')
        sample = self.sample_factory.make_sample(
            file_content, file_name,
            content_type, content_disposition)

        try:
            await self.db_con.analysis_add(sample)
        except PeekabooDatabaseError as dberr:
            logger.error('Failed to add analysis to database: %s', dberr)
            return sanic.response.json(
                {'message': 'Failed to add analysis to database'}, 500)

        if not self.job_queue.submit(sample):
            logger.error('Error submitting sample to job queue')
            return sanic.response.json(
                {'message': 'Error submitting sample to job queue'}, 500)

        # send answer to client
        return sanic.response.json({'job_id': sample.id}, 200)

    async def report(self, _, job_id):
        """ report endpoint for report retrieval by job ID

        @param request: sanic request object
        @type request: sanic.Request
        @param job_id: job ID extracted from endpoint path
        @type job_id: int
        @returns: report json response
        """
        if not job_id:
            return sanic.response.json(
                {'message': 'job ID missing from request'}, 400)

        try:
            job_info = await self.db_con.analysis_retrieve(job_id)
        except PeekabooDatabaseError as dberr:
            logger.error('Failed to retrieve analysis result from '
                         'database: %s', dberr)
            return sanic.response.json(
                {'message': 'Failed to retrieve analysis result '
                            'from database'}, 500)

        if job_info is None:
            logger.debug('No analysis result yet for job %d', job_id)
            return sanic.response.json(
                {'message': 'No analysis result yet for job %d' % job_id}, 404)

        reason, result = job_info
        return sanic.response.json({
            'result': result.name,
            'reason': reason,
            # FIXME: depends on saving the report to the database
            # 'report': report,
            }, 200)

    def serve(self):
        """ Serves requests until shutdown is requested from the outside. """
        self.server = self.loop.run_until_complete(self.server_coroutine)

        # sanic 21.9 introduced an explicit startup that finalizes the app,
        # particularly the request routing. So we need to run it if present.
        if hasattr(self.server, 'startup'):
            self.loop.run_until_complete(self.server.startup())

        self.loop.run_until_complete(self.server.start_serving())
        logger.info('Peekaboo server is now listening on %s:%d',
                    self.host, self.port)
        self.loop.run_until_complete(self.server.wait_closed())
        logger.debug('Server shut down.')

    def shut_down(self):
        """ Triggers a shutdown of the server, used by the signal handler and
        potentially other components to cause the main loop to exit. """
        logger.debug('Server shutdown requested.')
        if self.server is not None:
            self.server.close()



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# toolbox/                                                                    #
#         files.py                                                            #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2022 science + computing ag                              #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################

import logging
from datetime import datetime

from peekaboo.ruleset import Result


logger = logging.getLogger(__name__)


class Knowntools:
    """ Parent class, defines interface to knowntools. """
    def __init__(self, sample, db_con):
        self.sample = sample
        self.db_con = db_con

    def get_report(self):
        """ Return knowntools report or create if not already cached. """
        if self.sample.knowntools_report is not None:
            return self.sample.knowntools_report

        ktreport = KnowntoolsReport(
            self.db_con.analysis_journal_fetch_journal(self.sample)
        )

        self.sample.register_knowntools_report(ktreport)
        return ktreport


class KnowntoolsReport:
    """ Represents a custom Knowntools report. """
    def __init__(self, sample_journal=None):
        if sample_journal is None:
            sample_journal = []
        self.sample_journal = sample_journal

    def __str__(self):
        return "<KnowntoolsReport('%s'>" % self.sample_journal

    @property
    def known(self):
        """ Determines if there is a database entry for this sample """
        return bool(self.sample_journal)

    @property
    def last_result(self):
        """ Return the cached result """
        if self.sample_journal:
            return self.sample_journal[-1].result
        return Result.unknown

    @property
    def result(self):
        """ Return the cached result """
        return self.last_result

    def worst(self):
        """ Return the worst (result, reason) cached """
        worst_result = Result.unchecked
        worst_reason = ""
        for _, result, reason in self.sample_journal:
            if result > worst_result:
                worst_result = result
                worst_reason = reason
        return (worst_result, worst_reason)

    @property
    def worst_result(self):
        """ Return the worst result cached """
        return self.worst()[0]

    @property
    def first(self):
        """ Calculates the age in days since first record of this sample """
        if self.sample_journal:
            first = self.sample_journal[0].analysis_time
            now = datetime.today()
            difference = now - first
            return difference.days
        return 0

    @property
    def last(self):
        """ Calculates the age in days since most recent record of this
        sample """
        if self.sample_journal:
            last = self.sample_journal[-1].analysis_time
            now = datetime.today()
            difference = now - last
            return difference.days
        return 0



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# db.py                                                                       #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2022 science + computing ag                              #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################

""" A class wrapping database operations needed by Peekaboo based on
SQLAlchemy. """

import asyncio
import random
import time
import threading
import logging
from datetime import datetime, timedelta
from sqlalchemy import Column, Integer, String, Text, DateTime, \
        Enum, Index
import sqlalchemy.sql.expression
import sqlalchemy.ext.asyncio
import sqlalchemy.pool
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.engine import create_engine
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.exc import SQLAlchemyError, IntegrityError, OperationalError, \
        DBAPIError
from peekaboo import __version__
from peekaboo.ruleset import Result
from peekaboo.sample import JobState
from peekaboo.exceptions import PeekabooDatabaseError

DB_SCHEMA_VERSION = 9

logger = logging.getLogger(__name__)
Base = declarative_base()


#
# Database schema definition.
##############################################################################


class InFlightSample(Base):
    """
    Table tracking whether a specific sample is currently being analysed and by
    which Peekaboo instance.
    """
    __tablename__ = 'in_flight_samples_v%d' % DB_SCHEMA_VERSION

    # Indices:
    # - general considerations: The table will likely never have more than a
    #   couple of hundret entries. But we add and delete quite frequently.
    # - uniqueness of the primary key ensures atomic insertion when adding a
    #   lock
    # - column: we delete our own stale locks by instance_id.
    # - column: we delete other's stale locks by start_time.
    # - compound: we delete our own locks by sha256sum and instance_id.
    #   (admittedly a bit of overkill since the individual columns are already
    #   indexed.)

    sha256sum = Column(String(64), primary_key=True)
    instance_id = Column(Integer, nullable=False, index=True)
    start_time = Column(DateTime, nullable=False, index=True)

    __table_args__ = (
        # Index names need to be unique per schema in postgresql.
        Index('ix_%s_sha_iid' % __tablename__, sha256sum, instance_id),
        )

    def __str__(self):
        return (
            '<InFlightSample(sha256sum="%s", instance_id="%s", '
            'start_time="%s")>'
            % (self.sha256sum,
               self.instance_id,
               self.start_time.strftime("%Y%m%dT%H%M%S"))
        )

    __repr__ = __str__


class SampleInfo(Base):
    """ Definition of the sample_info table. """
    __tablename__ = 'sample_info_v%d' % DB_SCHEMA_VERSION

    # Indices:
    # - general considerations: The table grows very large over time. Every
    #   sample is checked against it to find a cached analysis result.
    #   Otherwise it's quite unused currently.
    # - compound: we fetch the analsysis journal by id, state, result,
    #   sha256sum and file extension

    id = Column(Integer, primary_key=True)
    state = Column(Enum(JobState), nullable=False)
    sha256sum = Column(String(64), nullable=False)
    file_extension = Column(String(16), nullable=True)
    analysis_time = Column(DateTime, nullable=False,
                           index=True)
    result = Column(Enum(Result), nullable=False)
    reason = Column(Text, nullable=True)

    __table_args__ = (
        Index('ix_%s_id_st_re_sha_fe' % __tablename__,
              id, state, result, sha256sum, file_extension),
    )

    def __str__(self):
        return ('<SampleInfo(sample_sha256_hash="%s", file_extension="%s", '
                'reason="%s", analysis_time="%s")>'
                % (self.sha256sum,
                   self.file_extension,
                   self.reason,
                   self.analysis_time.strftime("%Y%m%dT%H%M%S")))

    __repr__ = __str__


#
# End of database schema definition.
##############################################################################


class PeekabooDatabase:
    """ Peekaboo's database. """
    def __init__(self, db_url, instance_id=0,
                 stale_in_flight_threshold=15*60,
                 log_level=logging.WARNING,
                 async_driver=None):
        """
        Initialize the Peekaboo database handler.

        @param db_url: An RFC 1738 URL that points to the database.
        @param instance_id: A positive, unique ID differentiating this Peekaboo
                            instance from any other instance using the same
                            database for concurrency coordination. Value of 0
                            means that we're alone and have no other instances
                            to worry about.
        @param stale_in_flight_threshold: Number of seconds after which a in
        flight marker is considered stale and deleted or ignored.
        @param log_level: Overrides the log level of the database modules. The
                          idea is for the database to be silent by default and
                          only emit log messages if switched on explictly and
                          independently of the Peekaboo log level.
        @param async_driver: last resort override of the asyncio driver
                             auto-detection
        """
        logging.getLogger('sqlalchemy.engine').setLevel(log_level)
        logging.getLogger('sqlalchemy.pool').setLevel(log_level)
        # aiosqlite picks up the global log level unconditionally so we need to
        # override it as well and explicitly
        logging.getLogger('aiosqlite').setLevel(log_level)

        # <backend>[+<driver>]:// -> <backend>
        backend = db_url.split(':')[0].split('+')[0]

        connect_args = {}
        if backend == 'sqlite':
            connect_args['timeout'] = 0

        self.__engine = create_engine(
            db_url, future=True, connect_args=connect_args)
        session_factory = sessionmaker(bind=self.__engine)
        self.__session = scoped_session(session_factory)
        self.__lock = threading.RLock()

        asyncio_drivers = {
            'sqlite': ['aiosqlite'],
            'mysql': ['asyncmy', 'aiomysql'],
            'postgresql': ['asyncpg'],
        }

        if async_driver is not None:
            drivers = [async_driver]
        else:
            drivers = asyncio_drivers.get(backend)
            if drivers is None:
                raise PeekabooDatabaseError(
                    'Unknown database backend configured: %s' % backend)

        async_engine = None
        async_db_url = None
        for driver in drivers:
            # replace backend and driver with our asyncio alternative
            scheme = "%s+%s" % (backend, driver)
            async_db_url = ':'.join([scheme] + db_url.split(':')[1:])

            try:
                async_engine = sqlalchemy.ext.asyncio.create_async_engine(
                    async_db_url, connect_args=connect_args)
            except ModuleNotFoundError:
                continue

            logger.debug('Auto-detected %s SQLAlchemy backend+driver for '
                         'asyncio database accesses', scheme)
            break

        if async_engine is None:
            raise PeekabooDatabaseError(
                'None of the asyncio drivers for backend %s could be '
                'found: %s' % (backend, drivers))

        self.__async_session_factory = sessionmaker(
            bind=async_engine,
            class_=sqlalchemy.ext.asyncio.AsyncSession)
        # no scoping necessary as we're not using asyncio across threads

        # special handling for sqlite: since it does not respond well to
        # multiple modify operations in parallel to the same database, we
        # serialise them through a QueuePool with only one connection
        self.__async_session_factory_modify = self.__async_session_factory
        if backend in ['sqlite']:
            async_engine_modify = sqlalchemy.ext.asyncio.create_async_engine(
                async_db_url, poolclass=sqlalchemy.pool.AsyncAdaptedQueuePool,
                pool_size=1, max_overflow=0)
            self.__async_session_factory_modify = sessionmaker(
                bind=async_engine_modify,
                class_=sqlalchemy.ext.asyncio.AsyncSession)

        self.instance_id = instance_id
        self.stale_in_flight_threshold = stale_in_flight_threshold
        self.retries = 5
        # ultra-simple quadratic backoff:
        # attempt 1: 10 * 2**(1) == 10-20msecs
        # attempt 2: 10 * 2**(2) == 20-40msecs
        # attempt 3: 10 * 2**(3) == 40-80msecs
        # attempt 4: 10 * 2**(4) == 80-160msecs
        self.deadlock_backoff_base = 10
        self.connect_backoff_base = 2000

        attempt = 1
        delay = 0
        while attempt <= self.retries:
            with self.__lock:
                try:
                    Base.metadata.create_all(self.__engine)
                    break
                except (OperationalError, DBAPIError,
                        SQLAlchemyError) as error:
                    attempt, delay = self.was_transient_error(
                        error, attempt, 'create metadata')

                    if attempt < 0:
                        raise PeekabooDatabaseError(
                            'Failed to create schema in database: %s' % error)

            time.sleep(delay)

    def was_transient_error(self, error, attempt, action):
        """ Decide if an exception signals a transient error condition and
        sleep for some milliseconds if so.

        @param error: The exception object to look at.
        @type attempt: The current attempt number.
        @returns: The new attempt number or -1 if no further attempts should be
                  made.
        """
        # will not be retried anyway, so no use checking and sleeping
        if attempt >= self.retries:
            return -1, 0

        # only DBAPIError has connection_invalidated

        if getattr(error, 'connection_invalidated', False):
            logger.debug('Connection invalidated %s. Retrying.', action)
            return attempt + 1, 0

        # Access the original DBAPI exception anonymously.
        # We intentionally do some crude duck-typing here to avoid
        # imports of otherwise optional RDBMS modules. False-positive
        # would cause some useless retries of a different but
        # identically numbered error of another RDBMS.
        if (getattr(error, 'orig', None) is None or
                getattr(error.orig, 'args', None) is None):
            return -1, 0

        args = error.orig.args

        # (MySQLdb._exceptions.OperationalError) (2002, "Can't connect to local
        # MySQL server through socket '/var/run/mysqld/mysqld.sock' (2)")
        if (isinstance(args, tuple) and len(args) > 0 and args[0] in [2002, 2003]):
            # sleep some millisecs
            maxmsecs = self.connect_backoff_base * 2**attempt
            backoff = random.randint(maxmsecs/2, maxmsecs)
            logger.debug('Connection failed %s, backing off for %d '
                         'milliseconds before retrying', action, backoff)
            return attempt + 1, backoff / 1000

        # (MySQLdb._exceptions.OperationalError) (1213, 'Deadlock
        # found when trying to get lock; try restarting transaction')
        # (sqlite3.OperationalError) database is locked
        if (isinstance(args, tuple) and len(args) > 0 and
                args[0] in [1213, 'database is locked']):
            # sleep some millisecs
            maxmsecs = self.deadlock_backoff_base * 2**attempt
            backoff = random.randint(maxmsecs/2, maxmsecs)
            logger.debug('Database deadlock detected %s, backing off for %d '
                         'milliseconds before retrying.', action, backoff)
            return attempt + 1, backoff / 1000

        return -1, 0

    async def analysis_add(self, sample):
        """
        Add an analysis task to the analysis journal in the database.

        @param sample: The sample object for this analysis task.
        @returns: ID of the newly created analysis task (also updated
                  in the sample)
        """
        sample_info = SampleInfo(
            state=sample.state,
            sha256sum=sample.sha256sum,
            file_extension=sample.file_extension,
            analysis_time=datetime.now(),
            result=sample.result,
            reason=sample.reason)

        job_id = None
        attempt = 1
        delay = 0
        while attempt <= self.retries:
            async with self.__async_session_factory_modify() as session:
                session.add(sample_info)
                try:
                    # flush to retrieve the automatically assigned primary
                    # key value
                    await session.flush()
                    job_id = sample_info.id
                    await session.commit()
                    break
                except (OperationalError, DBAPIError,
                        SQLAlchemyError) as error:
                    await session.rollback()

                    attempt, delay = self.was_transient_error(
                       error, attempt, 'adding analysis')

                    if attempt < 0:
                        raise PeekabooDatabaseError(
                            'Failed to add analysis task to the database: %s' %
                            error)

            await asyncio.sleep(delay)

        sample.update_id(job_id)
        return job_id

    def analysis_update(self, sample):
        """
        Update an analysis task in the analysis journal in the database.

        @param sample: The sample object for this analysis task.
        """
        statement = sqlalchemy.sql.expression.update(SampleInfo).where(
            SampleInfo.id == sample.id).values(
                state=sample.state,
                result=sample.result,
                reason=sample.reason)

        attempt = 1
        delay = 0
        while attempt <= self.retries:
            with self.__lock:
                with self.__session() as session:
                    try:
                        session.execute(statement)
                        session.commit()
                        break
                    except (OperationalError, DBAPIError,
                            SQLAlchemyError) as error:
                        session.rollback()

                        attempt, delay = self.was_transient_error(
                            error, attempt, 'updating analysis')

                        if attempt < 0:
                            raise PeekabooDatabaseError(
                                'Failed to update analysis task in the database: %s' %
                                error)

            time.sleep(delay)

    def analysis_journal_fetch_journal(self, sample):
        """
        Fetch information stored in the database about a given sample object.

        @param sample: The sample object of which the information shall be
                       fetched from the database.
        @return: A sorted list of (analysis_time, result, reason) of the
                 requested sample.
        """
        statement = sqlalchemy.sql.expression.select(
            SampleInfo.analysis_time, SampleInfo.result,
            SampleInfo.reason).where(
                SampleInfo.id != sample.id).where(
                    SampleInfo.result != Result.failed).filter_by(
                        state=JobState.FINISHED,
                        sha256sum=sample.sha256sum,
                        file_extension=sample.file_extension).order_by(
                            SampleInfo.analysis_time)

        sample_journal = None
        attempt = 1
        delay = 0
        while attempt <= self.retries:
            with self.__session() as session:
                try:
                    sample_journal = session.execute(statement).all()
                    break
                except (OperationalError, DBAPIError,
                        SQLAlchemyError) as error:
                    session.rollback()

                    attempt, delay = self.was_transient_error(
                        error, attempt, 'fetching analysis journal')

                    if attempt < 0:
                        raise PeekabooDatabaseError(
                            'Failed to fetch analysis journal from the database: %s' %
                            error)

            time.sleep(delay)

        return sample_journal

    async def analysis_retrieve(self, job_id):
        """
        Fetch information stored in the database about a given sample object.

        @param job_id: ID of the analysis to retrieve
        @type job_id: int
        @return: reason and result for the given analysis task
        """
        statement = sqlalchemy.sql.expression.select(
            SampleInfo.reason, SampleInfo.result).filter_by(
                id=job_id, state=JobState.FINISHED)

        result = None
        attempt = 1
        delay = 0
        while attempt <= self.retries:
            async with self.__async_session_factory() as session:
                try:
                    proxy = await session.execute(statement)
                    result = proxy.first()
                    break
                except (OperationalError, DBAPIError,
                        SQLAlchemyError) as error:
                    await session.rollback()

                    attempt, delay = self.was_transient_error(
                        error, attempt, 'retrieving analysis result')

                    if attempt < 0:
                        raise PeekabooDatabaseError(
                            'Failed to retrieve analysis from the database: %s' %
                            error)

            await asyncio.sleep(delay)

        return result

    def mark_sample_in_flight(self, sample, instance_id=None, start_time=None):
        """
        Mark a sample as in flight, i.e. being worked on by an instance.

        @param sample: The sample to mark as in flight.
        @param instance_id: (optionally) The ID of the instance that is
                            handling this sample. Default: Us.
        @param start_time: Override the time the marker was placed for
                           debugging purposes.
        """
        # an instance id of 0 denotes that we're alone and don't need to track
        # in-flight samples in the database
        if self.instance_id == 0:
            return True

        # use our own instance id if none is given
        if instance_id is None:
            instance_id = self.instance_id

        if start_time is None:
            start_time = datetime.utcnow()

        in_flight_marker = InFlightSample(sha256sum=sample.sha256sum,
                                          instance_id=instance_id,
                                          start_time=start_time)
        attempt = 1
        delay = 0
        while attempt <= self.retries:
            # a new session needs to be constructed on each attempt
            with self.__session() as session:
                # try to mark this sample as in flight in an atomic insert
                # operation (modulo possible deadlocks with various RDBMS)
                session.add(in_flight_marker)

                try:
                    session.commit()
                    logger.debug('%d: Marked sample in flight', sample.id)
                    return True
                # duplicate primary key == entry already exists
                except IntegrityError:
                    session.rollback()
                    logger.debug('%d: Sample is already in flight on another '
                                 'instance', sample.id)
                    return False
                except (OperationalError, DBAPIError,
                        SQLAlchemyError) as error:
                    session.rollback()

                    attempt, delay = self.was_transient_error(
                        error, attempt, 'marking sample %d in flight' %
                        sample.id)

                    if attempt < 0:
                        raise PeekabooDatabaseError(
                            '%d: Unable to mark sample as in flight: %s' % (
                                sample.id, error))

            time.sleep(delay)

        return False

    def clear_sample_in_flight(self, sample, instance_id=None):
        """
        Clear the mark that a sample is being processed by an instance.

        @param sample: The sample to clear from in-flight list.
        @param instance_id: (optionally) The ID of the instance that is
                            handling this sample. Default: Us.
        """
        # an instance id of 0 denotes that we're alone and don't need to track
        # in-flight samples in the database
        if self.instance_id == 0:
            return

        # use our own instance id if none is given
        if instance_id is None:
            instance_id = self.instance_id

        statement = sqlalchemy.sql.expression.delete(
            InFlightSample).where(
                InFlightSample.sha256sum == sample.sha256sum).where(
                    InFlightSample.instance_id == instance_id)

        attempt = 1
        cleared = 0
        while attempt <= self.retries:
            with self.__session() as session:
                try:
                    # clear in-flight marker from database
                    marker = session.execute(statement)
                    session.commit()
                    cleared = marker.rowcount
                    break
                except (OperationalError, DBAPIError,
                        SQLAlchemyError) as error:
                    session.rollback()

                    attempt, delay = self.was_transient_error(
                        error, attempt, 'clearing in-flight status of '
                        'sample %d' % sample.id)

                    if attempt < 0:
                        raise PeekabooDatabaseError(
                            '%d: Unable to clear in-flight status of sample: '
                            '%s' % (sample.id, error))

            time.sleep(delay)

        if cleared == 0:
            raise PeekabooDatabaseError(
                '%d: Unexpected inconsistency: Sample not recorded as '
                'in-flight upon clearing flag.' % sample.id)
        elif cleared > 1:
            raise PeekabooDatabaseError(
                '%d: Unexpected inconsistency: Multiple instances of sample '
                'in-flight status cleared against database constraints!?' %
                sample.id)

    def clear_in_flight_samples(self, instance_id=None):
        """
        Clear all in-flight markers left over by previous runs or other
        instances by removing them from the lock table.

        @param instance_id: Clear our own (None), another instance's (positive
                            integer) or all instances' (negative integer) locks.
                            Since an instance_id of 0 disables in-flight sample
                            tracking, no instance will ever set a marker with
                            that ID so that specifying 0 here will amount to a
                            no-op or rather clean-up of invalid entries.
        """
        # an instance id of 0 denotes that we're alone and don't need to track
        # in-flight samples
        if self.instance_id == 0:
            return

        # use our own instance id if none is given
        if instance_id is None:
            instance_id = self.instance_id

        if instance_id < 0:
            # delete all locks
            statement = sqlalchemy.sql.expression.delete(InFlightSample)
            logger.debug('Clearing database of all in-flight samples.')
        else:
            # delete only the locks of a specific instance
            statement = sqlalchemy.sql.expression.delete(
                InFlightSample).where(
                    InFlightSample.instance_id == instance_id)
            logger.debug('Clearing database of all in-flight samples of '
                         'instance %d.', instance_id)

        attempt = 1
        while attempt <= self.retries:
            with self.__session() as session:
                try:
                    session.execute(statement)
                    session.commit()
                    break
                except (OperationalError, DBAPIError,
                        SQLAlchemyError) as error:
                    session.rollback()

                    attempt, delay = self.was_transient_error(
                        error, attempt,
                        'clearing database of in-flight samples')

                    if attempt < 0:
                        raise PeekabooDatabaseError(
                            'Unable to clear the database of in-flight '
                            'samples: %s' % error)

            time.sleep(delay)

    def clear_stale_in_flight_samples(self):
        """
        Clear all in-flight markers that are too old and therefore stale. This
        detects instances which are locked up, crashed or shut down.
        """
        # an instance id of 0 denotes that we're alone and don't need to track
        # in-flight samples in the database
        if self.instance_id == 0:
            return True

        logger.debug(
            'Clearing database of all stale in-flight samples '
            '(%d seconds)', self.stale_in_flight_threshold)

        def clear_statement(statement_class):
            # delete only the locks of a specific instance
            return statement_class(InFlightSample).where(
                InFlightSample.start_time <= datetime.utcnow() - timedelta(
                    seconds=self.stale_in_flight_threshold))

        delete_statement = clear_statement(sqlalchemy.sql.expression.delete)
        select_statement = clear_statement(sqlalchemy.sql.expression.select)

        attempt = 1
        cleared = 0
        while attempt <= self.retries:
            with self.__session() as session:
                try:
                    # only do the query if debugging is enabled
                    if logger.isEnabledFor(logging.DEBUG):
                        # obviously there's a race between logging and actual
                        # delete here, use with caution, compare with actual
                        # number of markers cleared below before relying on it
                        # for debugging
                        markers = session.execute(select_statement)
                        for stale in markers:
                            logger.debug(
                                'Stale in-flight marker to clear: %s', stale)

                    markers = session.execute(delete_statement)
                    session.commit()

                    cleared = markers.rowcount
                    if cleared > 0:
                        logger.warning(
                            '%d stale in-flight samples cleared.', cleared)

                    break
                except (OperationalError, DBAPIError,
                        SQLAlchemyError) as error:
                    session.rollback()

                    attempt, delay = self.was_transient_error(
                        error, attempt,
                        'clearing the database of stale in-flight samples')

                    if attempt < 0:
                        raise PeekabooDatabaseError(
                            'Unable to clear the database of stale in-flight '
                            'samples: %s' % error)

            time.sleep(delay)

        return cleared > 0

    def drop(self):
        """ Drop all tables of the database. """
        try:
            Base.metadata.drop_all(self.__engine)
        except SQLAlchemyError as error:
            raise PeekabooDatabaseError(
                'Unable to drop all tables of the database: %s' % error)
